"""
ModelPricer is an interface class for the parent pricing model that defines the following interfaces:
1. price_chain is the abstract method that is implemented in the parent class for a specific model and its params using analytic solution
2. model_mc_price_chain is the interface method for pricing option chain using mc of model dynamics
3. calibrate_model_params_to_chain is the interface method that using model based price_chain
the rest of interface methods are concrete relying on price_chain
market options data is passed using data container ChainData
"""

import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
import seaborn as sns
from numba.typed import List
from abc import ABC, abstractmethod
from scipy import stats
from dataclasses import dataclass
from typing import Tuple, Optional, Dict

from stochvolmodels.pricers.core.config import VariableType
from stochvolmodels.data.option_chain import OptionChain, OptionSlice
from stochvolmodels.utils import plots as plot
from stochvolmodels.utils.funcs import set_seed, update_kwargs

# set global mc seed, for resets call set_seed() locally
set_seed(24)


@dataclass
class ModelParams:
    pass


class ModelPricer(ABC):

    def __init__(self):
        super().__init__()

    #########################################################
    #                     generic interfaces
    #########################################################
    @abstractmethod
    def price_chain(self, option_chain: OptionChain, params: ModelParams, **kwargs
                    ) -> List[np.ndarray]:
        """
        abstract method for pricing chain data using model parameters
        recommended as a wrapper for numba implementation
        note that numba.List is equivalent type for Tuple
        """
        pass

    def compute_chain_prices_with_vols(self,
                                       option_chain: OptionChain,
                                       params: ModelParams,
                                       **kwargs
                                       ) -> Tuple[List[np.ndarray], List[np.ndarray]]:
        """
        price chain and compute model vols
        """
        model_prices = self.price_chain(option_chain=option_chain, params=params, **kwargs)
        model_ivols = option_chain.compute_model_ivols_from_chain_data(model_prices=model_prices)
        return model_prices, model_ivols

    def compute_model_ivols_for_chain(self,
                                      option_chain: OptionChain,
                                      params: ModelParams,
                                      **kwargs
                                      ) -> List[np.ndarray]:
        """
        price chain and compute model vols
        """
        model_prices, model_ivols = self.compute_chain_prices_with_vols(option_chain=option_chain,
                                                                        params=params,
                                                                        **kwargs)
        return model_ivols

    def model_mc_price_chain(self, option_chain: OptionChain, params: ModelParams, **kwargs
                             ) -> Tuple[List[np.ndarray], List[np.ndarray]]:
        """
        abstract method for pricing chain data using simulation of model dynamics
        recommended as a wrapper for numba implementation
        """
        raise NotImplementedError(f"must be implemented in parent class")

    def calibrate_model_params_to_chain(self, option_chain: OptionChain, **kwargs):
        """
        this is core method for model calibration
        we keep as not generic because model implementation may not require calibration of model parameters
        """
        raise NotImplementedError(f"must be implemented in parent class")

    #########################################################
    #          implemented interfaces for pricing
    #########################################################
    def price_slice(self,
                    params: ModelParams,
                    ttm: float,
                    forward: float,
                    strikes: np.ndarray,
                    optiontypes: np.ndarray,
                    discfactor: float = 1.0,
                    **kwargs
                    ) -> Tuple[np.ndarray, np.ndarray]:
        """
        price single slice_t using model chain pricer
        return price slices and model implied vols
        """
        option_chain = OptionChain.slice_to_chain(ttm=ttm,
                                                  forward=forward,
                                                  strikes=strikes,
                                                  optiontypes=optiontypes,
                                                  discfactor=discfactor)
        model_prices = self.price_chain(option_chain=option_chain,
                                        params=params,
                                        **kwargs)
        model_ivols = option_chain.compute_model_ivols_from_chain_data(model_prices=model_prices)
        return model_prices[0], model_ivols[0]

    def price_vanilla(self,
                      params: ModelParams,
                      ttm: float,
                      forward: float,
                      strike: float,
                      optiontype: str,
                      discfactor: float = 1.0,
                      **kwargs
                      ) -> Tuple[float, float]:
        """
        price single option using slice_t pricer
        return price and model implied vol
        """
        model_prices, model_ivols = self.price_slice(params=params, ttm=ttm, forward=forward,
                                                     strikes=np.array([strike]), optiontypes=np.array([optiontype]),
                                                     discfactor=discfactor, **kwargs)
        return model_prices[0], model_ivols[0]

    #########################################################
    #          implemented interfaces for monte carlo
    #########################################################
    def simulate_vol_paths(self, params: ModelParams, **kwargs) -> (np.ndarray, np.ndarray):
        """
        get grid of vol paths
        """
        raise NotImplementedError(f"must be implemented in parent class")

    def simulate_terminal_values(self, params: ModelParams, **kwargs) -> (np.ndarray, np.ndarray, np.ndarray):
        """
        get realizationss of terminal paths of x, vol, qvar
        """
        raise NotImplementedError(f"must be implemented in parent class")

    #########################################################
    #          implemented interfaces for mc implied vol comptutions
    #########################################################
    def compute_mc_chain_implied_vols(self,
                                      option_chain: OptionChain,
                                      params: ModelParams,
                                      nb_path: int = 100000,
                                      **kwargs
                                      ) -> Tuple[List[np.ndarray], ...]:
        """
        use model_mc_price_chain intefrace to compute model ivols with confidence bound
        """
        model_prices_ttms, option_std_ttms = self.model_mc_price_chain(option_chain=option_chain,
                                                                       params=params,
                                                                       nb_path=nb_path,
                                                                       **kwargs)
        std_factor = 1.96
        model_prices_ttms_ups = List()
        model_prices_ttms_downs = List()
        for model_prices_ttm, option_std_ttm in zip(model_prices_ttms, option_std_ttms):
            model_prices_ttms_ups.append(model_prices_ttm + std_factor * option_std_ttm)
            model_prices_ttms_downs.append(np.maximum(model_prices_ttm - std_factor * option_std_ttm, 1e-10))

        ivols_mid = option_chain.compute_model_ivols_from_chain_data(model_prices=model_prices_ttms)
        ivols_up = option_chain.compute_model_ivols_from_chain_data(model_prices=model_prices_ttms_ups)
        ivols_down = option_chain.compute_model_ivols_from_chain_data(model_prices=model_prices_ttms_downs)
        return model_prices_ttms, model_prices_ttms_ups, model_prices_ttms_downs, ivols_mid, ivols_up, ivols_down, option_std_ttms

    def get_log_return_mc_pdf(self,
                              ttm: float,
                              params: ModelParams,
                              x_grid: np.ndarray,
                              nb_path: int = 100000
                              ) -> np.ndarray:

        t_values = self.simulate_terminal_values(ttm=ttm, params=params, nb_path=nb_path)

        cut_off = 1e16
        # nan can be present too
        inf_nans = np.isnan(t_values)
        inf_pos = np.greater(t_values, cut_off, where=inf_nans == False)
        inf_neg = np.less(t_values, -cut_off, where=inf_nans == False)

        print(f"in mc: num -inf = {np.sum(inf_neg)}, num +inf = {np.sum(inf_pos)}, num nans = {np.sum(inf_nans)}")
        t_values = t_values[np.logical_and(np.logical_and(inf_neg == False, inf_pos == False), inf_nans == False)]
        kernel = stats.gaussian_kde(t_values)
        z = kernel(x_grid)
        # z = np.where(np.greater(z, 0.0), z, 0.0)
        z = z / np.nansum(z)
        return z

    #########################################################
    #          densities
    #########################################################
    def compute_logreturn_pdf(self, params: ModelParams, **kwargs) -> np.ndarray:
        """
        model pdf
        """
        raise NotImplementedError(f"must be implemented in parent class")

    #########################################################
    #          visualization interfaces
    #########################################################
    def plot_model_ivols(self,
                         option_chain: OptionChain,
                         params: ModelParams,
                         is_log_strike_xaxis: bool = False,
                         headers: Optional[List[str]] = None,
                         ax: plt.Subplot = None,
                         **kwargs
                         ) -> Optional[plt.Figure]:
        """
        plot model slice_t vols
        """
        model_ivols = self.compute_model_ivols_for_chain(option_chain=option_chain, params=params, **kwargs)

        if ax is None:
            with sns.axes_style('darkgrid'):
                fig, ax = plt.subplots(1, 1, figsize=plot.FIGSIZE, tight_layout=True)
        else:
            fig = None

        model_vols_ts = []
        for idx, ttm in enumerate(option_chain.ttms):
            if is_log_strike_xaxis:
                strikes = np.log(option_chain.strikes_ttms[idx] / option_chain.forwards[idx])
            else:
                strikes = option_chain.strikes_ttms[idx]

            if option_chain.ids is not None:
                if headers is not None:
                    name = f"{headers[idx]} slice - {option_chain.ids[idx]}"
                else:
                    name = f"Slice - {option_chain.ids[idx]}"
            else:
                name = f"{ttm=:0.2f}"
            model_vols_ts.append(pd.Series(model_ivols[idx], index=strikes, name=name))

        model_vols_ts = pd.concat(model_vols_ts, axis=1)

        plot.model_vols_ts(model_vols=model_vols_ts,
                           title='Model Implied Black Volatilities',
                           xlabel='log-strike' if is_log_strike_xaxis else 'strike',
                           xvar_format='{:0.2f}' if is_log_strike_xaxis else '{:0,.0f}',
                           x_rotation=0,
                           ax=ax,
                           **kwargs)
        return fig

    def plot_model_slices_in_params(self,
                                    option_slice: OptionSlice,
                                    params_dict: Dict[str, ModelParams],
                                    is_log_strike_xaxis: bool = False,
                                    ax: plt.Subplot = None,
                                    **kwargs
                                    ) -> Optional[plt.Figure]:
        """
        plot model slice_t vols
        """
        model_vols_pars = []
        for key, params in params_dict.items():
            prices, model_ivols = self.price_slice(ttm=option_slice.ttm,
                                                   forward=option_slice.forward,
                                                   strikes=option_slice.strikes,
                                                   optiontypes=option_slice.optiontypes,
                                                   params=params,
                                                   **kwargs)
            model_vols_pars.append(pd.Series(model_ivols, index=option_slice.strikes, name=key))

        model_vols_pars = pd.concat(model_vols_pars, axis=1)

        if ax is None:
            with sns.axes_style('darkgrid'):
                fig, ax = plt.subplots(1, 1, figsize=plot.FIGSIZE, tight_layout=True)
        else:
            fig = None

        plot.model_vols_ts(model_vols=model_vols_pars,
                           title='Model Vols',
                           xlabel='log-strike' if is_log_strike_xaxis else 'strike',
                           xvar_format='{:0.2f}' if is_log_strike_xaxis else '{:0,.0f}',
                           x_rotation=0,
                           ax=ax,
                           **kwargs)
        return fig

    def plot_model_ivols_vs_bid_ask(self,
                                    option_chain: OptionChain,
                                    params: ModelParams,
                                    is_log_strike_xaxis: bool = False,
                                    headers: Optional[List[str]] = None,
                                    **kwargs
                                    ) -> plt.Figure:
        """
        plot model slice_t vols
        optimized for 2*2 figure
        """
        model_ivols = self.compute_model_ivols_for_chain(option_chain=option_chain, params=params, **kwargs)

        with sns.axes_style('darkgrid'):
            fig, axs = plt.subplots(2, 2, figsize=plot.FIGSIZE, tight_layout=True)

        atm_vols = option_chain.get_chain_atm_vols()
        for idx, ttm in enumerate(option_chain.ttms):
            if is_log_strike_xaxis:
                strikes = np.log(option_chain.strikes_ttms[idx] / option_chain.forwards[idx])
                atm_forward = 0.0
                xvar_format = '{:0.2f}'
                strike_name = 'log-strike'
            else:
                strikes = option_chain.strikes_ttms[idx]
                atm_forward = option_chain.forwards[idx]
                xvar_format = '{:0,.0f}'
                strike_name = 'strike'

            midvols = 0.5 * (option_chain.bid_ivs[idx] + option_chain.ask_ivs[idx])
            mse2 = np.sqrt(np.nanmean(np.power(model_ivols[idx] - midvols, 2)))
            model_vols = pd.Series(model_ivols[idx], index=strikes, name=f"Model Fit: mse={mse2:0.2%}")
            if option_chain.ids is not None:
                if headers is not None:
                    title = f"{headers[idx]} slice - {option_chain.ids[idx]}"
                else:
                    title = f"Slice - {option_chain.ids[idx]}"
            else:
                title = f"{ttm=:0.2f}"

            if is_log_strike_xaxis:
                atm_points = {'ATM': (0.0, atm_vols[idx])}
            else:
                atm_points = {'ATM': (atm_forward, atm_vols[idx])}

            plot.vol_slice_fit(bid_vol=pd.Series(option_chain.bid_ivs[idx], index=strikes),
                               ask_vol=pd.Series(option_chain.ask_ivs[idx], index=strikes),
                               model_vols=model_vols,
                               title=title,
                               atm_points=atm_points,
                               strike_name=strike_name,
                               xvar_format=xvar_format,
                               x_rotation=0,
                               ax=axs[idx % 2][idx // 2],
                               **kwargs)
        return fig

    def plot_model_ivols_vs_mc(self,
                               option_chain: OptionChain,
                               params: ModelParams,
                               is_log_strike_xaxis: bool = False,
                               variable_type: VariableType = VariableType.LOG_RETURN,
                               nb_path: int = 100000,
                               **kwargs
                               ) -> None:
        """
        comparision of model implied vols computed old_analytics vs mc pricer
        optimized for 2*2 figure
        """
        model_ivols = self.compute_model_ivols_for_chain(option_chain=option_chain, params=params, **kwargs)

        mc_prices_ttms, _, _, mc_ivols, mc_ivols_up, mc_ivols_down, _ = self.compute_mc_chain_implied_vols(
            option_chain=option_chain,
            params=params,
            nb_path=nb_path,
            **kwargs)

        with sns.axes_style('darkgrid'):
            if len(option_chain.ttms) > 1:
                fig, axs = plt.subplots(2, len(option_chain.ttms) // 2, figsize=plot.FIGSIZE, tight_layout=True)
            else:
                fig, axs = plt.subplots(1, 1, figsize=plot.FIGSIZE, tight_layout=True)

        for idx, ttm in enumerate(option_chain.ttms):
            if is_log_strike_xaxis:
                strikes = np.log(option_chain.strikes_ttms[idx] / option_chain.forwards[idx])
                xvar_format = '{:0.2f}'
                strike_name = 'log-strike'
            else:
                strikes = option_chain.strikes_ttms[idx]
                if variable_type == VariableType.LOG_RETURN:
                    xvar_format = '{:0,.0f}'
                    strike_name = 'strike'
                else:
                    xvar_format = '{:0.2f}'
                    strike_name = 'QVAR strike'

            mse2 = np.sqrt(np.nanmean(np.power(model_ivols[idx] - mc_ivols[idx], 2)))
            model_vol_t = pd.Series(model_ivols[idx], index=strikes, name=f"Model: mse={mse2:0.2%}")

            if option_chain.ids is not None:
                title = f"{option_chain.ids[idx]}, {ttm=:0.2f}"
            else:
                title = f"{ttm=:0.2f}"

            if len(option_chain.ttms) > 1:
                ax = axs[idx % 2][idx // 2]
            else:
                ax = axs
            plot.vol_slice_fit(bid_vol=pd.Series(mc_ivols_down[idx], index=strikes),
                               ask_vol=pd.Series(mc_ivols_up[idx], index=strikes),
                               model_vols=model_vol_t,
                               title=title,
                               bid_name='MC: -0.95ci',
                               ask_name='MC: +0.95ci',
                               strike_name=strike_name,
                               xvar_format=xvar_format,
                               x_rotation=0,
                               ax=ax,
                               **kwargs)

    def plot_comp_mma_inverse_options_with_mc(self,
                                              option_chain: OptionChain,
                                              params: ModelParams,
                                              variable_type: VariableType = VariableType.LOG_RETURN,
                                              nb_path: int = 100000,
                                              headers: Optional[List[str]] = ('(A)', '(B)', '(C)', '(D)'),  # optimized for 2*2 figure
                                              is_log_strike_xaxis: bool = False,
                                              is_plot_vols: bool = True,
                                              figsize: Tuple[float, float] = plot.FIGSIZE,
                                              **kwargs
                                              ) -> plt.Figure:
        """
        comparision of model implied vols computed old_analytics under MMA and inverse measures vs mc pricer
        optimized for 2*2 figure
        """
        mma_label = 'MMA'
        inverse_lable = 'Inverse'

        model_prices_mma, model_ivols_mma = self.compute_chain_prices_with_vols(option_chain=option_chain, params=params,
                                                                                is_spot_measure=True,
                                                                                variable_type=variable_type,
                                                                                **kwargs)

        model_prices_inv, model_ivols_inv= self.compute_chain_prices_with_vols(option_chain=option_chain, params=params,
                                                                               is_spot_measure=False,
                                                                               variable_type=variable_type,
                                                                               **kwargs)

        # we perform MC simulation in MMA measure
        mc_kwargs = update_kwargs(kwargs, dict(is_spot_measure=True, variable_type=variable_type))
        model_prices_ttms, model_prices_ttms_ups, model_prices_ttms_downs, \
        mc_ivols, mc_ivols_up, mc_ivols_down, mc_stdev_ttms = self.compute_mc_chain_implied_vols(option_chain=option_chain,
                                                                                                 params=params,
                                                                                                 nb_path=nb_path,
                                                                                                 **mc_kwargs)

        if is_plot_vols:
            model_datas = {mma_label: model_ivols_mma, inverse_lable: model_ivols_inv}
            mc_data = mc_ivols
            mc_data_lower, mc_data_upper = mc_ivols_down, mc_ivols_up
        else:
            model_datas = {mma_label: model_prices_mma, inverse_lable: model_prices_inv}
            mc_data = model_prices_ttms
            mc_data_lower, mc_data_upper = model_prices_ttms_downs, model_prices_ttms_ups

        if option_chain.ttms.size < 4:
            nrows, ncols = 1, option_chain.ttms.size
        else:
            nrows, ncols = 2, option_chain.ttms.size//2

        with sns.axes_style('darkgrid'):
            fig, axs = plt.subplots(nrows, ncols, figsize=figsize, tight_layout=True)

        for idx, ttm in enumerate(option_chain.ttms):
            if is_log_strike_xaxis:
                strikes = np.log(option_chain.strikes_ttms[idx] / option_chain.forwards[idx])
                xvar_format = '{:0.2f}'
                strike_name = 'log-strike'
            else:
                strikes = option_chain.strikes_ttms[idx]
                if variable_type == VariableType.LOG_RETURN:
                    xvar_format = '{:0,.2f}'
                    strike_name = 'strike'
                else:
                    xvar_format = '{:0.2f}'
                    strike_name = 'QVAR strike'

            model_vols = {}
            for key, model_data in model_datas.items():
                mse = np.sqrt(np.nanmean(np.power(model_data[idx] - mc_data[idx], 2)))
                model_vols[f"{key}: mse={mse:0.2%}"] = pd.Series(model_data[idx], index=strikes)
            model_vols = pd.DataFrame.from_dict(model_vols, orient='columns')

            if option_chain.ids is not None:
                if headers is not None:
                    title = f"{headers[idx]} slice - {option_chain.ids[idx]}"
                else:
                    title = f"slice - {option_chain.ids[idx]}"
            else:
                title = f"{ttm=:0.2f}"

            atm_vol = np.interp(x=option_chain.forwards[idx], xp=option_chain.strikes_ttms[idx],
                                fp=0.5 * (mc_data_lower[idx] + mc_data_upper[idx]))
            if is_log_strike_xaxis:
                atm_points = {'ATM': (0.0, atm_vol)}
            else:
                atm_points = {'ATM': (option_chain.forwards[idx], atm_vol)}

            plot.vol_slice_fit(bid_vol=pd.Series(mc_data_lower[idx], index=strikes),
                               ask_vol=pd.Series(mc_data_upper[idx], index=strikes),
                               model_vols=model_vols,
                               title=title,
                               bid_name='MC: -0.95ci',
                               ask_name='MC: +0.95ci',
                               strike_name=strike_name,
                               xvar_format=xvar_format,
                               x_rotation=0,
                               atm_points=atm_points,
                               ylabel='Implied vols' if is_plot_vols else 'Model prices',
                               yvar_format='{:.0%}' if is_plot_vols else '{:.2f}',
                               ax=axs[idx] if axs.ndim == 1 else axs[idx % 2][idx // 2],
                               **kwargs)
        return fig




# built in
import numpy as np
import matplotlib.pyplot as plt
from scipy.optimize import minimize, fsolve
from numba.typed import List
from typing import Tuple, Optional, Dict, Any
from dataclasses import dataclass, asdict
from scipy.integrate import solve_ivp
from scipy.integrate._ivp.ivp import OdeResult
from enum import Enum

# stochvolmodels pricers
from stochvolmodels.pricers.core import mgf_pricer as mgfp
from stochvolmodels.pricers.core.config import VariableType
from stochvolmodels.pricers.core.mc_payoffs import compute_mc_vars_payoff
from stochvolmodels.pricers.model_pricer import ModelPricer, ModelParams
from stochvolmodels.utils.funcs import to_flat_np_array, set_time_grid, timer, set_seed

# data
from stochvolmodels.data.option_chain import OptionChain
from stochvolmodels.data.test_option_chain import get_btc_test_chain_data


@dataclass
class HawkesJDParams(ModelParams):
    """
    parameters of 2-factor Hawkes Jump Diffusion
    annualized params, close for BTC on daily frequency
    """
    mu: float = 0.0
    sigma: float = 0.45
    # jumps
    shift_p: float = 0.06
    mean_p: float = 0.03
    shift_m: float = -0.06
    mean_m: float = -0.03
    # positive jumps intensity
    lambda_p: float = 6.55
    theta_p: float = 6.55
    kappa_p: float = 22.29
    beta1_p: float = 76.0
    beta2_p: float = -67.58#*0.0
    # minus jumps intensity
    lambda_m: float = 8.50
    theta_m: float = 8.50
    kappa_m: float = 29.0
    beta1_m: float = 104.55#*0.0
    beta2_m: float = -109.6

    def __post_init__(self):
        self.compensator_p = np.exp(self.shift_p)/(1.0-self.mean_p) - 1.0
        self.compensator_m = np.exp(self.shift_m)/(1.0-self.mean_m) - 1.0

    def to_dict(self) -> Dict[str, Any]:
        return asdict(self)

    def print(self) -> None:
        for k, v in self.to_dict().items():
            print(f"{k}={v}")
        print('condifions')
        print(f"jump1={self.jump1_cond:0.4f} > 0")
        print(f"jump2={self.jump2_cond:0.4f} > 0")

    @property
    def jump1_cond(self) -> float:
        return self.kappa_p-self.beta1_p*self.exp_jump_p-self.beta2_p*self.exp_jump_m

    @property
    def jump2_cond(self) -> float:
        return self.kappa_m - self.beta2_m * self.exp_jump_m - self.beta1_m * self.exp_jump_p

    @property
    def exp_jump_p(self) -> float:
        return self.shift_p+self.mean_p

    @property
    def exp_jump_m(self) -> float:
        return self.shift_m+self.mean_m

    @property
    def jumps_var_m(self) -> float:
        return np.square(self.shift_m) + np.square(self.mean_m)

    @property
    def jumps_var_p(self) -> float:
        return np.square(self.shift_p) + np.square(self.mean_p)


@dataclass
class RiskPremiaGammas:
    x: float = 0.0
    lambda_p: float = 0.0
    lambda_m: float = 0.0


class HawkesJDPricer(ModelPricer):

    @timer
    def price_chain(self,
                    option_chain: OptionChain,
                    params: HawkesJDParams,
                    is_spot_measure: bool = True,
                    risk_premia_gammas: RiskPremiaGammas = None,
                    **kwargs
                    ) -> List[np.ndarray]:
        """
        implementation of generic method price_chain using log sv wrapper
        """
        if risk_premia_gammas is not None:
            model_prices = hawkesjd_chain_pricer_with_risk_premia(model_params=params,
                                                                  risk_premia_gammas=risk_premia_gammas,
                                                                  ttms=option_chain.ttms,
                                                                  forwards=option_chain.forwards,
                                                                  discfactors=option_chain.discfactors,
                                                                  strikes_ttms=option_chain.strikes_ttms,
                                                                  optiontypes_ttms=option_chain.optiontypes_ttms,
                                                                  is_spot_measure=is_spot_measure,
                                                                  **kwargs)
        else:
            model_prices = hawkesjd_chain_pricer(model_params=params,
                                                 ttms=option_chain.ttms,
                                                 forwards=option_chain.forwards,
                                                 discfactors=option_chain.discfactors,
                                                 strikes_ttms=option_chain.strikes_ttms,
                                                 optiontypes_ttms=option_chain.optiontypes_ttms,
                                                 is_spot_measure=is_spot_measure,
                                                 **kwargs)
        return model_prices


    @timer
    def model_mc_price_chain(self,
                             option_chain: OptionChain,
                             params: HawkesJDParams,
                             nb_path: int = 100000,
                             **kwargs
                             ) -> (List[np.ndarray], List[np.ndarray]):
        return hawkesjd_mc_chain_pricer(ttms=option_chain.ttms,
                                        forwards=option_chain.forwards,
                                        discfactors=option_chain.discfactors,
                                        strikes_ttms=option_chain.strikes_ttms,
                                        optiontypes_ttms=option_chain.optiontypes_ttms,
                                        nb_path=nb_path,
                                        **params.to_dict())

    @timer
    def calibrate_model_params_to_chain(self,
                                        option_chain: OptionChain,
                                        params0: HawkesJDParams,
                                        is_vega_weighted: bool = True,
                                        is_unit_ttm_vega: bool = False,
                                        **kwargs
                                        ) -> HawkesJDParams:
        """
        implementation of model calibration interface
        fit: sigma, mean_p, mean_m, theta_p, theta_m, kappa, (=kappa_p, kappa_m), beta_p (=beta1_p, beta2_p), beta_m (=beta1_m, beta2_m)
        use fixed shift_p, shift_m, lambda_p, lambda_m
        """
        x, y = option_chain.get_chain_data_as_xy()
        market_vols = to_flat_np_array(y)  # market mid quotes
        if is_vega_weighted:
            vegas_ttms = option_chain.get_chain_vegas(is_unit_ttm_vega=is_unit_ttm_vega)
            vegas_ttms = [vegas_ttm/sum(vegas_ttm) for vegas_ttm in vegas_ttms]
            weights = to_flat_np_array(vegas_ttms)
        else:
            weights = np.ones_like(market_vols)

        p0 = np.array([params0.sigma, params0.mean_p, params0.mean_m, params0.theta_p, params0.theta_m,
                       0.5*(params0.kappa_p+params0.kappa_m),
                       0.5*(params0.beta1_p-params0.beta2_p),
                       0.5*(params0.beta2_p-params0.beta2_m)])
        bounds = ((0.10, 2.0), (0.01, 0.99), (-0.99, -0.01), (0.01, 100.0), (0.01, 100.0),
                  (1.0, 100.0), (1.0, 100.0), (1.0, 100.0))

        def unpack_pars(pars: np.ndarray) -> HawkesJDParams:
            sigma, mean_p, mean_m, theta_p, theta_m, kappa, beta_p, beta_m \
                = pars[0], pars[1], pars[2], pars[3], pars[4], pars[5], pars[6], pars[7]
            model_params = HawkesJDParams(mu=0.0,
                                          sigma=sigma,
                                          shift_p=params0.shift_p,
                                          mean_p=mean_p,
                                          shift_m=params0.shift_m,
                                          mean_m=mean_m,
                                          lambda_p=params0.lambda_p,
                                          theta_p=theta_p,
                                          kappa_p=kappa,
                                          beta1_p=beta_p,
                                          beta2_p=-beta_p,
                                          lambda_m=params0.lambda_m,
                                          theta_m=theta_m,
                                          kappa_m=kappa,
                                          beta1_m=beta_m,
                                          beta2_m=-beta_m)
            return model_params

        def objective(pars: np.ndarray, args: np.ndarray) -> float:
            params = unpack_pars(pars=pars)
            model_vols = self.compute_model_ivols_for_chain(option_chain=option_chain, params=params)
            resid = np.nansum(weights * np.square(to_flat_np_array(model_vols) - market_vols))
            return resid

        def jump_cond(pars: np.ndarray) -> float:
            params = unpack_pars(pars=pars)
            return params.jump1_cond + params.jump2_cond

        constraints = ({'type': 'ineq', 'fun': jump_cond})
        options = {'disp': True, 'ftol': 1e-8}

        if constraints is not None:
            res = minimize(objective, p0, args=None, method='SLSQP', constraints=constraints, bounds=bounds, options=options)
        else:
            res = minimize(objective, p0, args=None, method='SLSQP', bounds=bounds, options=options)

        fit_params = unpack_pars(pars=res.x)
        return fit_params


def set_vol_scaler(sigma0: float, ttm: float) -> float:
    return np.minimum(sigma0, 0.3) * np.sqrt(np.minimum(ttm, 0.5 / 12.0))  # lower bound is two w


def hawkesjd_chain_pricer(model_params: HawkesJDParams,
                          ttms: np.ndarray,
                          forwards: np.ndarray,
                          discfactors: np.ndarray,
                          strikes_ttms: List[np.ndarray],
                          optiontypes_ttms: List[np.ndarray],
                          is_stiff_solver: bool = False,
                          is_spot_measure: bool = True,
                          variable_type: VariableType = VariableType.LOG_RETURN,
                          vol_scaler: float = None
                          ) -> List[np.ndarray]:
    """
    wrapper to price option chain on variable_type
    to do: do numba implementation using numba consistent solver
    """
    # starting values
    if vol_scaler is None:  # for calibrations we fix one vol_scaler so the grid is not affected by v0
        vol_scaler = set_vol_scaler(sigma0=model_params.sigma, ttm=np.min(ttms))

    phi_grid, psi_grid, theta_grid = mgfp.get_transform_var_grid(variable_type=variable_type,
                                                                 is_spot_measure=is_spot_measure,
                                                                 vol_scaler=vol_scaler)
    a_t0 = np.zeros((phi_grid.shape[0], 3), dtype=np.complex128)
    ttm0 = 0.0

    # outputs as numpy lists
    model_prices_ttms = List()
    for ttm, forward, strikes_ttm, optiontypes_ttm, discfactor in zip(ttms, forwards, strikes_ttms,
                                                                      optiontypes_ttms, discfactors):

        a_t0, log_mgf_grid = compute_hawkes_a_mgf_grid(ttm=ttm - ttm0,
                                                       phi_grid=phi_grid,
                                                       psi_grid=psi_grid,
                                                       theta_grid=theta_grid,
                                                       a_t0=a_t0,
                                                       is_stiff_solver=is_stiff_solver,
                                                       model_params=model_params)

        if variable_type == VariableType.LOG_RETURN:
            option_prices = mgfp.slice_pricer_with_mgf_grid(log_mgf_grid=log_mgf_grid,
                                                            phi_grid=phi_grid,
                                                            ttm=ttm,
                                                            forward=forward,
                                                            strikes=strikes_ttm,
                                                            optiontypes=optiontypes_ttm,
                                                            discfactor=discfactor,
                                                            is_spot_measure=is_spot_measure)
        else:
            raise NotImplementedError

        model_prices_ttms.append(option_prices)
        ttm0 = ttm

    return model_prices_ttms


def hawkesjd_chain_pricer_with_risk_premia(model_params: HawkesJDParams,
                                           risk_premia_gammas: RiskPremiaGammas,
                                           ttms: np.ndarray,
                                           forwards: np.ndarray,
                                           discfactors: np.ndarray,
                                           strikes_ttms: List[np.ndarray],
                                           optiontypes_ttms: List[np.ndarray],
                                           is_stiff_solver: bool = False,
                                           is_spot_measure: bool = True,
                                           variable_type: VariableType = VariableType.LOG_RETURN,
                                           vol_scaler: float = None
                                           ) -> List[np.ndarray]:
    """
    wrapper to price option chain on variable_type
    to do: do numba implementation using numba consistent solver
    """
    print('risk_premia_gammas')
    # starting values
    if vol_scaler is None:  # for calibrations we fix one vol_scaler so the grid is not affected by v0
        vol_scaler = set_vol_scaler(sigma0=model_params.sigma, ttm=np.min(ttms))

    phi_grid, psi_grid, theta_grid = mgfp.get_transform_var_grid(variable_type=variable_type,
                                                                 is_spot_measure=is_spot_measure,
                                                                 vol_scaler=vol_scaler,
                                                                 real_phi=-0.5-risk_premia_gammas.x)
    a_t0 = np.zeros((phi_grid.shape[0], 3), dtype=np.complex128)
    a_t0[:, 1] = risk_premia_gammas.lambda_p
    a_t0[:, 2] = risk_premia_gammas.lambda_m

    ttm0 = 0.0

    # outputs as numpy lists
    model_prices_ttms = List()
    for ttm, forward, strikes_ttm, optiontypes_ttm, discfactor in zip(ttms, forwards, strikes_ttms,
                                                                      optiontypes_ttms, discfactors):

        a_t1_, log_mgf1_ = compute_hawkes_a_mgf_grid(phi_grid=np.array([risk_premia_gammas.x]), ttm=ttm,
                                                     model_params=model_params,
                                                     a1_t0=a_t0[0, :],
                                                     is_stiff_solver=True)
        normalizer = np.exp(np.real(log_mgf1_))[0]
        print(normalizer)

        a_t0, log_mgf_grid = compute_hawkes_a_mgf_grid(ttm=ttm - ttm0,
                                                       phi_grid=phi_grid,
                                                       psi_grid=psi_grid,
                                                       theta_grid=theta_grid,
                                                       a_t0=a_t0,
                                                       is_stiff_solver=is_stiff_solver,
                                                       model_params=model_params)

        if variable_type == VariableType.LOG_RETURN:
            option_prices = mgfp.slice_pricer_with_mgf_grid(log_mgf_grid=log_mgf_grid,
                                                            phi_grid=phi_grid,
                                                            ttm=ttm,
                                                            forward=forward,
                                                            strikes=strikes_ttm,
                                                            optiontypes=optiontypes_ttm,
                                                            discfactor=discfactor,
                                                            is_spot_measure=is_spot_measure,
                                                            normalizer=normalizer)
        else:
            raise NotImplementedError

        model_prices_ttms.append(option_prices)
        ttm0 = ttm

    return model_prices_ttms


def compute_hawkes_a_mgf_grid(ttm: float,
                              phi_grid: np.ndarray,
                              model_params: HawkesJDParams,
                              psi_grid: Optional[np.ndarray] = None,  # Q-var grid
                              a_t0: Optional[np.ndarray] = None,
                              is_stiff_solver: bool = False,
                              **kwargs
                              ) -> Tuple[np.ndarray, np.ndarray]:
    """
    compute
     1. solution a_t1 for ode A given a_t0
     2. log mgf function: we save an exponent calulation when pricing options
    mmg in x or QV as function of phi
    ode_solution is computed per grid of phi
    to do: numba implementation: need numba ode solvers
    """

    if a_t0 is None:
        a_t0 = np.zeros((phi_grid.shape[0], 3), dtype=np.complex128)

    a_t1 = solve_a_ode_grid(phi_grid=phi_grid,
                            psi_grid=psi_grid,
                            ttm=ttm,
                            model_params=model_params,
                            a_t0=a_t0,
                            is_stiff_solver=is_stiff_solver)

    log_mgf = a_t1[:, 0] + a_t1[:, 1]*model_params.lambda_p + a_t1[:, 2]*model_params.lambda_m
    return a_t1, log_mgf


# cannot use @njit(cache=False, fastmath=True) when using solve_ode_for_a with solve_ivp
def solve_a_ode_grid(phi_grid: np.ndarray,  # x grid
                     ttm: float,
                     model_params: HawkesJDParams,
                     psi_grid: Optional[np.ndarray] = None,  # Q-var grid
                     a_t0: Optional[np.ndarray] = None,
                     is_stiff_solver: bool = False
                     ) -> np.ndarray:
    """
    solve ode for range phi
    next: numba implementation to compute in range of phi
    """
    if a_t0 is None:
        a_t0 = np.zeros((phi_grid.shape[0], 3), dtype=np.complex128)

    if psi_grid is None:
        psi_grid = np.zeros_like(phi_grid)

    f = lambda phi, psi, a0_: solve_ode_for_a(ttm=ttm,
                                              model_params=model_params,
                                              phi=phi,
                                              psi=psi,
                                              a_t0=a0_,
                                              is_stiff_solver=is_stiff_solver,
                                              dense_output=False)

    a_t1 = np.zeros((phi_grid.shape[0], 3), dtype=np.complex128)
    for idx, (phi, psi) in enumerate(zip(phi_grid, psi_grid)):
        a_t1[idx, :] = f(phi, psi, a_t0[idx, :]).y[:, -1]
    return a_t1


# cannot use @njit(cache=False, fastmath=True) when using solve_ode_for_a with solve_ivp
def solve_ode_for_a(ttm: float,
                    model_params: HawkesJDParams,
                    phi: np.complex128,
                    psi: np.complex128,
                    a_t0: Optional[np.ndarray] = None,
                    is_stiff_solver: bool = False,
                    dense_output: bool = False
                    ) -> OdeResult:
    """
    solve ode for given phi
    next: numba implementation to compute in range of phi
    """
    def e_p(phi_: float):
        return np.exp(-model_params.shift_p*phi_) / (1.0+model_params.mean_p*phi_)

    def e_m(phi_: float):
        return np.exp(-model_params.shift_m*phi_) / (1.0+model_params.mean_m*phi_)

    def func_rhs(t: float, #  dummy for ode solve
                 a0: np.ndarray
                 ) -> np.ndarray:
        rhs = np.zeros(3, dtype=np.complex128)

        j_p = e_p(phi_=phi - model_params.beta1_p * a0[1] - model_params.beta1_m * a0[2]) - 1.0
        j_m = e_m(phi_=phi - model_params.beta2_p * a0[1] - model_params.beta2_m * a0[2]) - 1.0

        rhs[0] = model_params.kappa_p*model_params.theta_p*a0[1] + model_params.kappa_m*model_params.theta_m*a0[2] \
                 + np.square(model_params.sigma)*(0.5*(phi+1.0)*phi-psi)
        rhs[1] = j_p - model_params.kappa_p*a0[1] + model_params.compensator_p*phi
        rhs[2] = j_m - model_params.kappa_m*a0[2] + model_params.compensator_m*phi
        return rhs

    if a_t0 is None:
        a_t0 = np.zeros(3, dtype=np.complex128)

    if is_stiff_solver:
        ode_sol = solve_ivp(fun=func_rhs, t_span=(0.0, ttm), y0=a_t0, args=None,
                            method='BDF',
                            # jac=func_rhs_jac,
                            dense_output=dense_output)
    else:
        ode_sol = solve_ivp(fun=func_rhs, t_span=(0.0, ttm), y0=a_t0, args=None,
                            dense_output=dense_output)

    return ode_sol


#@njit(cache=False, fastmath=True)
def hawkesjd_mc_chain_pricer(ttms: np.ndarray,
                             forwards: np.ndarray,
                             discfactors: np.ndarray,
                             strikes_ttms: Tuple[np.ndarray,...],
                             optiontypes_ttms: Tuple[np.ndarray, ...],
                             lambda_p: float,
                             lambda_m: float,
                             mu: float,
                             sigma: float,
                             shift_p: float,
                             mean_p: float,
                             shift_m: float,
                             mean_m: float,
                             theta_p: float,
                             kappa_p: float,
                             beta1_p: float,
                             beta2_p: float,
                             theta_m: float,
                             kappa_m: float,
                             beta1_m: float,
                             beta2_m: float,
                             nb_path: int = 100000,
                             variable_type: VariableType = VariableType.LOG_RETURN
                             ) -> Tuple[List[np.ndarray], List[np.ndarray]]:

    # starting values
    x0 = np.zeros(nb_path)
    lambda_p0 = lambda_p*np.ones(nb_path)
    lambda_m0 = lambda_m*np.ones(nb_path)
    ttm0 = 0.0

    # outputs as numpy lists
    option_prices_ttm = List()
    option_std_ttm = List()
    for ttm, forward, discfactor, strikes_ttm, optiontypes_ttm in zip(ttms, forwards, discfactors, strikes_ttms, optiontypes_ttms):
        x0, lambda_p0, lambda_m0 = simulate_hawkesjd_terminal(ttm=ttm - ttm0,
                                                              x0=x0,
                                                              lambda_p0=lambda_p0,
                                                              lambda_m0=lambda_m0,
                                                              mu=mu,
                                                              sigma=sigma,
                                                              shift_p=shift_p,
                                                              mean_p=mean_p,
                                                              shift_m=shift_m,
                                                              mean_m=mean_m,
                                                              theta_p=theta_p,
                                                              kappa_p=kappa_p,
                                                              beta1_p=beta1_p,
                                                              beta2_p=beta2_p,
                                                              theta_m=theta_m,
                                                              kappa_m=kappa_m,
                                                              beta1_m=beta1_m,
                                                              beta2_m=beta2_m,
                                                              nb_path=nb_path)
        ttm0 = ttm
        option_prices, option_std = compute_mc_vars_payoff(x0=x0, sigma0=x0, qvar0=x0,
                                                           ttm=ttm,
                                                           forward=forward,
                                                           strikes_ttm=strikes_ttm,
                                                           optiontypes_ttm=optiontypes_ttm,
                                                           discfactor=discfactor,
                                                           variable_type=variable_type)
        option_prices_ttm.append(option_prices)
        option_std_ttm.append(option_std)

    return option_prices_ttm, option_std_ttm


#@njit(cache=False, fastmath=True)
def simulate_hawkesjd_terminal(ttm: float,
                               x0:  np.ndarray,
                               lambda_p0: np.ndarray,
                               lambda_m0: np.ndarray,
                               mu: float,
                               sigma: float,
                               shift_p: float,
                               mean_p: float,
                               shift_m: float,
                               mean_m: float,
                               theta_p: float,
                               kappa_p: float,
                               beta1_p: float,
                               beta2_p: float,
                               theta_m: float,
                               kappa_m: float,
                               beta1_m: float,
                               beta2_m: float,
                               nb_path: int = 100000
                               ) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:

    if x0.shape[0] == 1:  # initial value
        x0 = x0*np.zeros(nb_path)
    else:
        assert x0.shape[0] == nb_path
    if lambda_p0.shape[0] == 1:  # initial value
        lambda_p0 = lambda_p0*np.ones(nb_path)
    else:
        assert lambda_p0.shape[0] == nb_path
    if lambda_m0.shape[0] == 1:  # initial value
        lambda_m0 = lambda_m0*np.ones(nb_path)
    else:
        assert lambda_m0.shape[0] == nb_path

    # vars
    nb_steps, dt, grid_t = set_time_grid(ttm=ttm, year_days=5*360)  # need small dt step for large intensities
    W0 = np.sqrt(dt) * np.random.normal(0, 1, size=(nb_steps, nb_path))
    U_P = -np.log(np.random.uniform(low=1e-16, high=1.0, size=(nb_steps, nb_path)))/dt
    U_M = -np.log(np.random.uniform(low=1e-16, high=1.0, size=(nb_steps, nb_path)))/dt
    J_P = shift_p + np.random.exponential(scale=mean_p, size=(nb_steps, nb_path))
    J_M = shift_m - np.random.exponential(scale=-mean_m, size=(nb_steps, nb_path))

    # params
    compensator_p_dt = dt*(np.exp(shift_p) / (1.0 - mean_p) - 1.0)
    compensator_m_dt = dt*(np.exp(shift_m) / (1.0 - mean_m) - 1.0)

    drift_dt = (mu-0.5*sigma*sigma) * dt
    for t_, (w0, u_p, u_m, j_p, j_m) in enumerate(zip(W0, U_P, U_M, J_P, J_M)):
        diffusion = drift_dt - compensator_p_dt*lambda_p0 - compensator_m_dt*lambda_m0 + sigma * w0
        # generate jumps:
        jump_p = np.where(lambda_p0 > u_p, j_p, 0.0)
        jump_m = np.where(lambda_m0 > u_m, j_m, 0.0)
        x0 = x0 + diffusion + jump_p + jump_m
        load_p = beta1_p*jump_p + beta2_p*jump_m
        load_m = beta1_m*jump_p + beta2_m*jump_m
        lambda_p0 = lambda_p0 + kappa_p*(theta_p-lambda_p0)*dt + load_p
        lambda_m0 = lambda_m0 + kappa_m*(theta_m-lambda_m0)*dt + load_m

    return x0, lambda_p0, lambda_m0


def solve_esscher_transform(params: HawkesJDParams,
                            ttm: float,
                            risk_premia_gammas: RiskPremiaGammas,
                            target_x: float = 0.0
                            ) -> RiskPremiaGammas:
    # these are fied for phis
    a1_t0 = np.zeros((2, 3), dtype=np.complex128)
    a1_t0[0, 0], a1_t0[0, 1], a1_t0[0, 2] = 0.0, risk_premia_gammas.lambda_p, risk_premia_gammas.lambda_m
    a1_t0[1, 0], a1_t0[1, 1], a1_t0[1, 2] = 0.0, risk_premia_gammas.lambda_p, risk_premia_gammas.lambda_m

    def func(x: float) -> float:
        phi_grid = -1.0*np.array([x, x+1.0])
        a_t1, log_mgf1 = compute_hawkes_a_mgf_grid(phi_grid=phi_grid, ttm=ttm, model_params=params, a1_t0=a1_t0, is_stiff_solver=True)
        return np.real(log_mgf1[1]-log_mgf1[0]) - target_x
    gamma = fsolve(func, x0=np.array([0.0]))
    risk_premia_gammas.x = gamma[0]
    return risk_premia_gammas


class UnitTests(Enum):
    OPTION_PRICER = 1
    CHAIN_PRICER = 2
    SLICE_PRICER = 3
    MC_COMPARISION = 4
    CALIBRATOR = 5


def run_unit_test(unit_test: UnitTests):

    params = HawkesJDParams(sigma=0.2,
                            shift_p=0.0,
                            mean_p=0.2,
                            shift_m=0.0,
                            mean_m=-0.1,
                            lambda_p=2.0,
                            theta_p=2.0,
                            kappa_p=50.0,
                            beta1_p=100.0,
                            beta2_p=0.0,
                            lambda_m=2.0,
                            theta_m=2.0,
                            kappa_m=50.0,
                            beta1_m=0.0,
                            beta2_m=-100.0)

    params = HawkesJDParams()

    params.print()
    pricer = HawkesJDPricer()

    set_seed(3)
    np.random.seed(3)

    if unit_test == UnitTests.OPTION_PRICER:

        model_price, vol = pricer.price_vanilla(params=params,
                                                ttm=0.25,
                                                forward=100.0,
                                                strike=100.0,
                                                optiontype='C')
        print(f"price={model_price:0.4f}, implied vol={vol: 0.2%}")

    elif unit_test == UnitTests.CHAIN_PRICER:
        option_chain = get_btc_test_chain_data()
        # option_chain = OptionChain.get_uniform_chain(flat_vol=params.sigma)
        model_prices = pricer.price_chain(option_chain=option_chain, params=params)
        print(model_prices)
        pricer.plot_model_ivols_vs_bid_ask(option_chain=option_chain, params=params)

        option_chain = OptionChain.to_uniform_strikes(option_chain, num_strikes=31)
        pricer.plot_model_ivols(option_chain=option_chain,
                                params=params)

        # pricer.plot_model_ivols_vs_mc(option_chain=option_chain, params=params, nb_path=400000)

    if unit_test == UnitTests.SLICE_PRICER:
        ttm = 1.0
        forward = 1.0
        strikes = np.array([0.9, 1.0, 1.1])
        optiontypes = np.array(['P', 'C', 'C'])

        model_prices, vols = pricer.price_slice(params=params,
                                                ttm=ttm,
                                                forward=forward,
                                                strikes=strikes,
                                                optiontypes=optiontypes)
        print(model_prices)
        print(vols)

        for strike, optiontype in zip(strikes, optiontypes):
            model_price, vol = pricer.price_vanilla(params=params,
                                                    ttm=ttm,
                                                    forward=forward,
                                                    strike=strike,
                                                    optiontype=optiontype)
            print(f"{model_price}, {vol}")

    elif unit_test == UnitTests.MC_COMPARISION:
        option_chain = get_btc_test_chain_data()
        # option_chain = OptionChain.get_uniform_chain(ttms=np.array([0.25]), ids=np.array(['3m']), strikes=100.0*np.linspace(0.5, 2.0, 15))

        pricer.plot_model_ivols_vs_mc(option_chain=option_chain,
                                      params=params,
                                      nb_path=100000)

    elif unit_test == UnitTests.CALIBRATOR:
        option_chain = get_btc_test_chain_data()
        fit_params = pricer.calibrate_model_params_to_chain(option_chain=option_chain,
                                                            params0=params)
        print('calibrated params')
        fit_params.print()
        pricer.plot_model_ivols_vs_bid_ask(option_chain=option_chain,
                                           params=fit_params)

    plt.show()


if __name__ == '__main__':

    unit_test = UnitTests.MC_COMPARISION

    is_run_all_tests = False
    if is_run_all_tests:
        for unit_test in UnitTests:
            run_unit_test(unit_test=unit_test)
    else:
        run_unit_test(unit_test=unit_test)



"""
config file
"""

from enum import Enum


class VariableType(Enum):
    """
    state variables for log SV model
    """
    LOG_RETURN = 1  # with transform var PHI
    Q_VAR = 2  # with transform var PSI
    SIGMA = 3  # with trasform for THETA



"""
key analytics for Black Scholes Merton pricer and implied volatilities
"""

import numpy as np
from numba import njit
from typing import Union, Tuple
from enum import Enum
from numba.typed import List
from scipy.stats import norm

from stochvolmodels.utils.funcs import ncdf, npdf


class OptionType(str, Enum):
    CALL = 'C'
    PUT = 'P'
    INVERSE_CALL = 'IC'
    INVERSE_PUT = 'IP'


@njit(cache=False, fastmath=True)
def compute_bsm_price(forward: float,
                      strike: float,
                      ttm: float,
                      vol: float,
                      optiontype: str = 'C',
                      discfactor: float = 1.0
                      ) -> float:
    """
    bsm pricer for forward
    """
    sT = vol * np.sqrt(ttm)
    d1 = (np.log(forward / strike) + 0.5 * sT * sT) / sT
    d2 = d1 - sT
    if optiontype == 'C' or optiontype == 'IC':
        price = discfactor * (forward * ncdf(d1) - strike * ncdf(d2))
    elif optiontype == 'P' or optiontype == 'IP':
        price = -discfactor * (forward * ncdf(-d1) - strike * ncdf(-d2))
    else:
        raise NotImplementedError(f"optiontype")

    return price


@njit(cache=False, fastmath=True)
def compute_bsm_slice_prices(ttm: float,
                             forward: float,
                             strikes: np.ndarray,
                             vols: np.ndarray,
                             optiontypes: np.ndarray,
                             discfactor: float = 1.0
                             ) -> np.ndarray:
    """
    vectorised bsm deltas for array of aligned strikes, vols, and optiontypes
    """
    def f(strike: float, vol: float, optiontype: str) -> float:
        return compute_bsm_price(forward=forward,
                                 ttm=ttm,
                                 vol=vol,
                                 strike=strike,
                                 optiontype=optiontype,
                                 discfactor=discfactor)

    bsm_prices = np.zeros_like(strikes)
    for idx, (strike, vol, optiontype) in enumerate(zip(strikes, vols, optiontypes)):
        bsm_prices[idx] = f(strike, vol, optiontype)
    return bsm_prices


def compute_bsm_delta_to_strike(ttm: float,
                             forward: float,
                             delta: float,
                             vol: float
                             ) -> Union[float, np.ndarray]:
    """
    bsm deltas for strikes and vols
    """
    inv_delta = norm.ppf(delta) if delta > 0.0 else -norm.ppf(-delta)
    sT = vol * np.sqrt(ttm)
    strike = forward*np.exp(-sT*(inv_delta - 0.5 * sT))
    return strike


@njit(cache=False, fastmath=True)
def compute_bsm_delta(ttm: float,
                      forward: float,
                      strike: float,
                      vol: float,
                      optiontype: str,
                      ) -> float:
    """
    bsm deltas for strikes and vols
    """
    sT = vol * np.sqrt(ttm)
    d1 = np.log(forward / strike) / sT + 0.5 * sT
    if optiontype == 'C':
        d1_sign = 1.0
    elif optiontype == 'P':
        d1_sign = - 1.0
    else:
        d1_sign = 0.0
    bsm_deltas = d1_sign * ncdf(d1_sign * d1)
    return bsm_deltas


@njit(cache=False, fastmath=True)
def compute_bsm_slice_deltas(ttm: Union[float, np.ndarray],
                             forward: Union[float, np.ndarray],
                             strikes: Union[float, np.ndarray],
                             vols: Union[float, np.ndarray],
                             optiontypes: Union[str, np.ndarray]
                             ) -> Union[float, np.ndarray]:
    """
    bsm deltas for strikes and vols
    """
    sT = vols * np.sqrt(ttm)
    d1 = np.log(forward / strikes) / sT + 0.5 * sT
    if optiontypes == 'C':
        d1_sign = 1.0
    elif optiontypes == 'P':
        d1_sign = - 1.0
    else:
        d1_sign = np.where(np.array([op == 'C' for op in optiontypes]), 1.0, -1.0)
    bsm_deltas = d1_sign * ncdf(d1_sign * d1)
    return bsm_deltas


#@njit(cache=False, fastmath=True)
def compute_bsm_deltas_ttms(ttms: np.ndarray,
                            forwards: np.ndarray,
                            strikes_ttms: Tuple[np.ndarray, ...],
                            vols_ttms: Tuple[np.ndarray,...],
                            optiontypes_ttms: Tuple[np.ndarray, ...],
                            ) -> List[np.ndarray]:
    """
    vectorised bsm deltas for array of aligned strikes, vols, and optiontypes
    """
    deltas_ttms = List()
    for ttm, forward, vols_ttm, strikes_ttm, optiontypes_ttm in zip(ttms, forwards, vols_ttms, strikes_ttms, optiontypes_ttms):
        deltas_ttms.append(compute_bsm_slice_deltas(ttm=ttm, forward=forward, strikes=strikes_ttm, vols=vols_ttm, optiontypes=optiontypes_ttm))
    return deltas_ttms


@njit(cache=False, fastmath=True)
def compute_bsm_slice_vegas(ttm: float,
                            forward: float,
                            strikes: np.ndarray,
                            vols: np.ndarray,
                            optiontypes: np.ndarray = None
                            ) -> np.ndarray:
    """
    vectorised bsm vegas for array of aligned strikes, vols, and optiontypes
    """
    sT = vols * np.sqrt(ttm)
    d1 = np.log(forward / strikes) / sT + 0.5 * sT
    vegas = forward * npdf(d1) * np.sqrt(ttm)
    return vegas


@njit(cache=False, fastmath=True)
def compute_bsm_vegas_ttms(ttms: np.ndarray,
                           forwards: np.ndarray,
                           strikes_ttms: Tuple[np.ndarray, ...],
                           vols_ttms: Tuple[np.ndarray,...],
                           optiontypes_ttms: Tuple[np.ndarray, ...],
                           ) -> List[np.ndarray]:
    """
    vectorised bsm vegas for array of aligned strikes, vols, and optiontypes
    """
    vegas_ttms = List()
    for ttm, forward, vols_ttm, strikes_ttm, optiontypes_ttm in zip(ttms, forwards, vols_ttms, strikes_ttms, optiontypes_ttms):
        vegas_ttms.append(compute_bsm_slice_vegas(ttm=ttm, forward=forward, strikes=strikes_ttm, vols=vols_ttm, optiontypes=optiontypes_ttm))
    return vegas_ttms


@njit(cache=False, fastmath=True)
def infer_bsm_ivols_from_model_slice_prices(ttm: float,
                                            forward: float,
                                            strikes: np.ndarray,
                                            optiontypes: np.ndarray,
                                            model_prices: np.ndarray,
                                            discfactor: float
                                            ) -> np.ndarray:
    model_vol_ttm = np.zeros_like(strikes)
    for idx, (strike, model_price, optiontype) in enumerate(zip(strikes, model_prices, optiontypes)):
        model_vol_ttm[idx] = infer_bsm_implied_vol(forward=forward, ttm=ttm, discfactor=discfactor,
                                                   given_price=model_price,
                                                   strike=strike,
                                                   optiontype=optiontype)
    return model_vol_ttm


@njit(cache=False, fastmath=True)
def infer_bsm_implied_vol(forward: float,
                          ttm: float,
                          strike: float,
                          given_price: float,
                          discfactor: float = 1.0,
                          optiontype: str = 'C',
                          tol: float = 1e-8,
                          is_bounds_to_nan: bool = True
                          ) -> float:
    """
    compute Black implied vol
    """
    x1, x2 = 0.01, 5.0  # starting values
    f = compute_bsm_price(forward=forward, strike=strike, ttm=ttm, vol=x1, discfactor=discfactor, optiontype=optiontype) - given_price
    fmid = compute_bsm_price(forward=forward, strike=strike, ttm=ttm, vol=x2, discfactor=discfactor, optiontype=optiontype) - given_price

    if f*fmid < 0.0:
        if f < 0.0:
            rtb = x1
            dx = x2-x1
        else:
            rtb = x2
            dx = x1-x2
        xmid = rtb
        for j in range(0, 40):
            dx = dx*0.5
            xmid = rtb+dx
            fmid = compute_bsm_price(forward=forward, strike=strike, ttm=ttm, vol=xmid, discfactor=discfactor, optiontype=optiontype) - given_price
            if fmid <= 0.0:
                rtb = xmid
            if np.abs(fmid) < tol:
                break
        v1 = xmid

    else:
        if f < 0:
            v1 = x1
        else:
            v1 = x2

    if is_bounds_to_nan:  # in case vol was inferred it will return nan
        if np.abs(v1-x1) < tol or np.abs(v1-x2) < tol:
            v1 = np.nan
    return v1


@njit(cache=False, fastmath=True)
def infer_bsm_ivols_from_slice_prices(ttm: float,
                                      forward: float,
                                      discfactor: float,
                                      strikes_ttm: np.ndarray,
                                      optiontypes_ttm: np.ndarray,
                                      model_prices_ttm: np.ndarray,
                                      ) -> np.ndarray:
    """
    vectorised chain ivols
    """
    model_vol_ttm = np.zeros_like(strikes_ttm)
    for idx, (strike, model_price, optiontype) in enumerate(zip(strikes_ttm, model_prices_ttm, optiontypes_ttm)):
        model_vol_ttm[idx] = infer_bsm_implied_vol(forward=forward, ttm=ttm, discfactor=discfactor,
                                                   given_price=model_price,
                                                   strike=strike,
                                                   optiontype=optiontype)
    return model_vol_ttm


@njit(cache=False, fastmath=True)
def infer_bsm_ivols_from_model_chain_prices(ttms: np.ndarray,
                                            forwards: np.ndarray,
                                            discfactors: np.ndarray,
                                            strikes_ttms: List[np.ndarray, ...],
                                            optiontypes_ttms: List[np.ndarray, ...],
                                            model_prices_ttms: List[np.ndarray],
                                            ) -> List[np.ndarray, ...]:
    """
    vectorised chain ivols
    """
    model_vol_ttms = List()
    for ttm, forward, discfactor, strikes_ttm, optiontypes_ttm, model_prices_ttm in zip(ttms, forwards, discfactors, strikes_ttms, optiontypes_ttms, model_prices_ttms):
        model_vol_ttm = np.zeros_like(strikes_ttm)
        for idx, (strike, model_price, optiontype) in enumerate(zip(strikes_ttm, model_prices_ttm, optiontypes_ttm)):
            model_vol_ttm[idx] = infer_bsm_implied_vol(forward=forward, ttm=ttm, discfactor=discfactor,
                                                       given_price=model_price,
                                                       strike=strike,
                                                       optiontype=optiontype)
        model_vol_ttms.append(model_vol_ttm)
    return model_vol_ttms






"""
generic analytics for option pricing using Fourier transforms of payoffs
"""

import numpy as np
from numba import njit
from typing import Tuple

from ...pricers.core.config import VariableType


@njit(cache=False, fastmath=True)
def get_phi_grid(is_spot_measure: bool = True,
                 vol_scaler: float = 0.28
                 ) -> np.ndarray:
    """
    for x = log-price variable
    vol_scaler = sigma_0*sqrt(ttm) will adjust the grid size: smaller val need longer period
    default vol scaler corresponds to pricing option with vol=100% and 1m ttm = 1/12
    """
    p = np.linspace(0, 5.6/vol_scaler, 1000)  # default max phi is 20 with 1000 points
    if is_spot_measure:
        real_p = -0.5
    else:
        real_p = 0.5
    phi_grid = real_p + 1j * p
    return phi_grid


@njit(cache=False, fastmath=True)
def get_psi_grid() -> np.ndarray:
    """
    for I = QV variable
    """
    p = np.linspace(0, 200, 4000)
    real_p = -0.5
    psi_grid = real_p + 1j * p
    return psi_grid


@njit(cache=False, fastmath=True)
def get_theta_grid() -> np.ndarray:
    """
    for sigma
    """
    p = np.linspace(0, 600, 4000)
    real_p = -0.5
    theta_grid = real_p + 1j * p
    return theta_grid


@njit(cache=False, fastmath=True)
def get_transform_var_grid(variable_type: VariableType = VariableType.LOG_RETURN,
                           is_spot_measure: bool = True,
                           vol_scaler: float = 0.28
                           ) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    """
    compute grid for Fourier inversions
    """
    if variable_type == VariableType.LOG_RETURN:
        phi_grid = get_phi_grid(is_spot_measure=is_spot_measure, vol_scaler=vol_scaler)
        psi_grid = np.zeros_like(phi_grid, dtype=np.complex128)
        theta_grid = np.zeros_like(phi_grid, dtype=np.complex128)

    elif variable_type == VariableType.Q_VAR:
        psi_grid = get_psi_grid()
        if is_spot_measure:
            phi_grid = np.zeros_like(psi_grid, dtype=np.complex128)
        else:
            phi_grid = np.ones_like(psi_grid, dtype=np.complex128)
        theta_grid = np.zeros_like(phi_grid, dtype=np.complex128)

    elif variable_type == VariableType.SIGMA:
        theta_grid = get_theta_grid()
        phi_grid = np.zeros_like(theta_grid, dtype=np.complex128)
        psi_grid = np.zeros_like(theta_grid, dtype=np.complex128)

    else:
        raise NotImplementedError
    return phi_grid, psi_grid, theta_grid


@njit(cache=False, fastmath=True)
def compute_integration_weights(var_grid: np.ndarray,
                                is_simpson: bool = True
                                ) -> np.ndarray:
    p = np.imag(var_grid)
    if is_simpson:
        dp = 2.0*np.ones(len(p))
        dp[0] = 1.0
        dp[-1] = 1.0
        for i in range(len(p)):
            if i % 2 == 1:
                dp[i] = 4.0
        dp = ((p[1] - p[0])/3.0) * dp
    else:  # trapezoidal rule
        dp = np.append(0.5*(p[1] - p[0]), p[1:] - p[:-1])
    return dp


@njit(cache=False, fastmath=True)
def slice_pricer_with_mgf_grid(log_mgf_grid: np.ndarray,
                               phi_grid: np.ndarray,
                               ttm: float,
                               forward: float,
                               strikes: np.ndarray,
                               optiontypes: np.ndarray,
                               discfactor: float = 1.0,
                               is_spot_measure: bool = True,
                               is_simpson: bool = True
                               ) -> np.ndarray:
    """
    generic function for pricing options on the spot given the mgf grid
    mgf in x is function defined on log-price transform phi grids
    transform variable is phi_grid = real_phi + i*p
    grid can be non-uniform
    """
    p = np.imag(phi_grid)
    dp = compute_integration_weights(var_grid=phi_grid, is_simpson=is_simpson)

    if np.all(np.abs(np.real(phi_grid))-0.5 < 1e-10):  # optimized for phi = +/-0.5 + i*p
        p_payoff = (dp / np.pi) / (p * p + 0.25) + 1j * 0.0  # add zero complex part for numba
    else:
        if is_spot_measure:
            p_payoff = - (dp / np.pi) / ((phi_grid+1.0)*phi_grid)
        else:
            p_payoff = - (dp / np.pi) / ((phi_grid-1.0) * phi_grid)

    log_strikes = np.log(forward/strikes)
    option_prices = np.zeros_like(log_strikes)
    for idx, (x, strike, type_) in enumerate(zip(log_strikes, strikes, optiontypes)):
        # compute sum using trapesoidal rule
        capped_option_price = np.nansum(np.real(p_payoff*np.exp(-x * phi_grid + log_mgf_grid)))
        if is_spot_measure:
            if type_ == 'C':
                option_prices[idx] = discfactor*(forward - strike * capped_option_price)
            elif type_ == 'P':
                option_prices[idx] = discfactor*(strike - strike * capped_option_price)
            else:
                raise ValueError(f"not implemented")
        else:  # for inverse measure we multiply by forward
            if type_ in ['IC', 'C']:
                option_prices[idx] = forward*discfactor*(1.0 - capped_option_price)
            elif type_ in ['IP', 'P']:
                option_prices[idx] = forward*discfactor*(np.exp(-x) - capped_option_price)
            else:
                raise ValueError(f"not implemented")

    return option_prices


@njit(cache=False, fastmath=True)
def slice_qvar_pricer_with_a_grid(log_mgf_grid: np.ndarray,
                                  psi_grid: np.ndarray,
                                  ttm: float,
                                  strikes: np.ndarray,
                                  optiontypes: np.ndarray,
                                  forward: float,
                                  discfactor: float = 1.0,
                                  is_simpson: bool = True,
                                  is_spot_measure: bool = True
                                  ) -> np.ndarray:
    """
    generic pricer of options on quadratic var using mgf
    mmg in x as function of a and phi grids
    ode_solution is computed per grid of phi
    """
    option_prices = np.zeros_like(strikes)
    log_strikes = np.log(forward / strikes)
    dp = compute_integration_weights(var_grid=psi_grid, is_simpson=is_simpson)
    p_payoff = (dp / np.pi) / (psi_grid * psi_grid)
    for idx, (log_strike, strike, type_) in enumerate(zip(log_strikes, strikes, optiontypes)):
        option_price = np.nansum(np.real(p_payoff * np.exp((strike * ttm) * psi_grid + log_mgf_grid)))
        option_price = np.maximum(discfactor * option_price / ttm, 1e-10)

        if is_spot_measure:
            if type_ == 'C':
                option_prices[idx] = option_price
            else:
                raise ValueError(f"not implemented")
        else:
            if type_ == 'C':
                option_prices[idx] = option_price
            else:
                raise ValueError(f"not implemented")

    return option_prices


@njit(cache=False, fastmath=True)
def pdf_with_mgf_grid(log_mgf_grid: np.ndarray,
                      transform_var_grid: np.ndarray,
                      space_grid: np.ndarray,
                      shift: float = 0.0,
                      is_simpson: bool = True
                      ) -> np.ndarray:
    """
    generic function for pricing options on the spot given the mgf grid
    mgf in x is function defined on log-price transform phi grids
    transform variable is phi_grid = real_phi + i*p
    grid can be non-uniform
    """
    dp = compute_integration_weights(var_grid=transform_var_grid, is_simpson=is_simpson) / np.pi
    pdf = np.zeros_like(space_grid)
    for idx, x in enumerate(space_grid):
        pdf[idx] = np.nansum(np.real(dp * np.exp((x-shift) * transform_var_grid + log_mgf_grid)))
    dx = space_grid[1] - space_grid[0]
    pdf = dx * pdf
    return pdf



"""
utility functions
"""
import functools
import time
import numpy as np
import pandas as pd
from numba import njit
from numba.core.decorators import njit
from numba.typed import List
from typing import Tuple, Dict, Any, Optional, Union


def to_flat_np_array(input_list: List[np.ndarray]) -> np.ndarray:
    return np.concatenate(input_list).ravel()


@njit(cache=False, fastmath=True)
def set_time_grid(ttm: float, year_days: float = 360) -> Tuple[int, float, np.ndarray]:
    """
    set daily steps
    """
    nb_steps = int(np.ceil(year_days * ttm))
    grid_t = np.linspace(0.0, ttm, nb_steps + 1)
    dt = grid_t[1] - grid_t[0]
    return nb_steps, dt, grid_t


@njit(cache=False, fastmath=True)
def set_seed(value):
    """
    set seed for numba space
    """
    np.random.seed(value)


def timer(func):
    """Print the runtime of the decorated function"""
    @functools.wraps(func)
    def wrapper_timer(*args, **kwargs):
        start_time = time.perf_counter()    # 1
        value = func(*args, **kwargs)
        end_time = time.perf_counter()      # 2
        run_time = end_time - start_time    # 3
        print(f"Finished {func.__name__!r} in {run_time:.4f} secs")
        return value
    return wrapper_timer


def compute_histogram_data(data: np.ndarray,
                           x_grid: np.ndarray,
                           name: str = 'Histogram'
                           ) -> pd.Series:
    """
    compute histogram on defined discrete grid
    """
    hist_data, bin_edges = np.histogram(a=data,
                                        bins=len(x_grid)-1,
                                        range=(x_grid[0], x_grid[-1]))
    hist_data = np.append(np.array(x_grid[0]), hist_data)
    hist_data = hist_data / len(data)
    hist_data = pd.Series(hist_data, index=bin_edges, name=name)
    return hist_data


def update_kwargs(kwargs: Dict[Any, Any],
                  new_kwargs: Optional[Dict[Any, Any]]
                  ) -> Dict[Any, Any]:
    """
    update kwargs with optional kwargs dicts
    """
    local_kwargs = kwargs.copy()
    if new_kwargs is not None and not len(new_kwargs) == 0:
        local_kwargs.update(new_kwargs)
    return local_kwargs


@njit(cache=False, fastmath=True)
def erfcc(x: Union[float, np.ndarray]) -> Union[float, np.ndarray]:
    """
    Complementary error function. can be vectorized
    """
    z = np.abs(x)
    t = 1. / (1. + 0.5*z)
    r = t * np.exp(-z*z-1.26551223+t*(1.00002368+t*(0.37409196+t*(0.09678418+t*(-0.18628806+t*(0.27886807+
        t*(-1.13520398+t*(1.48851587+t*(-.82215223+t*0.17087277)))))))))
    fcc = np.where(np.greater(x, 0.0), r, 2.0-r)
    return fcc


@njit(cache=False, fastmath=True)
def ncdf(x: Union[float, np.ndarray]) -> Union[float, np.ndarray]:
    return 1. - 0.5*erfcc(x/(np.sqrt(2.0)))


@njit(cache=False, fastmath=True)
def npdf(x: Union[float, np.ndarray]) -> Union[float, np.ndarray]:
    return np.exp(-0.5*np.square(x))/np.sqrt(2.0*np.pi)



"""
numba analytics for affine expansion
"""

import numpy as np
import numpy.linalg as la
from numba import njit
from enum import Enum
from typing import Tuple, Optional
from scipy.integrate import solve_ivp
from scipy.integrate._ivp.ivp import OdeResult

from ...pricers.core.config import VariableType


class ExpansionOrder(Enum):
    FIRST = 1
    SECOND = 2


@njit(cache=False, fastmath=True)
def get_expansion_n(expansion_order: ExpansionOrder = ExpansionOrder.FIRST) -> int:
    if expansion_order == ExpansionOrder.FIRST:
        n = 3
    else:
        n = 5
    return n


@njit(cache=False, fastmath=True)
def func_a_ode_quadratic_terms(theta: float,
                               kappa1: float,
                               kappa2: float,
                               beta: float,
                               volvol: float,
                               phi: np.complex128,
                               psi: np.complex128,
                               is_spot_measure: bool = True,
                               expansion_order: ExpansionOrder = ExpansionOrder.FIRST
                               ) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    """
    Matrices for the quadratic form A_t = A.T@M@A + L@A + H
    """
    theta2 = theta * theta
    vartheta2 = beta * beta + volvol * volvol
    qv = theta * vartheta2
    qv2 = theta2 * vartheta2
    if is_spot_measure:
        lamda = 0
        kappa_p = kappa1 + kappa2 * theta
        kappa2_p = kappa2
    else:
        lamda = beta*theta2
        kappa_p = kappa1 + kappa2 * theta - 2*beta*theta
        kappa2_p = kappa2-beta

    # fill Ms: M should be of same type as L and H for numba, eventhough they are real
    # utilize that M is symmetric
    # although tuple is more intuitive for M, because of numba we need to make it a tensor
    n = get_expansion_n(expansion_order=expansion_order)
    M = np.zeros((n, n, n), dtype=np.complex128)
    M[0, 1, 1] = 0.5 * qv2

    M[1, 1, 1] = qv
    M[1, 1, 2] = M[1, 2, 1] = qv2

    M[2, 1, 1], M[2, 2, 2] = 0.5 * vartheta2, 2.0 * qv2
    M[2, 2, 1] = M[2, 1, 2] = 2.0 * qv

    if expansion_order == ExpansionOrder.SECOND:
        M[2, 1, 3] = M[2, 3, 1] = 1.5*qv2
        M[3, 2, 2] = 4.0*qv
        M[3, 1, 2] = M[3, 2, 1] = vartheta2
        M[3, 1, 3] = M[3, 3, 1] = 3.0*qv
        M[3, 1, 4] = M[3, 4, 1] = 2.0 * qv2
        M[3, 2, 3] = M[3, 3, 2] = 3.0 * qv2

        M[4, 2, 2], M[4, 3, 3] = 2.0 * vartheta2, 4.5*qv2
        M[4, 1, 3] = M[4, 3, 1] = 1.5 * vartheta2
        M[4, 1, 4] = M[4, 4, 1] = 4.0 * qv
        M[4, 2, 3] = M[4, 3, 2] = 6.0 * qv
        M[4, 2, 4] = M[4, 4, 2] = 4.0 * qv2

    # fills Ls
    L = np.zeros((n, n), dtype=np.complex128)
    L[0, 1], L[0, 2] = lamda - theta2 * beta * phi, qv2
    L[1, 1], L[1, 2] = -kappa_p - 2.0 * theta * beta * phi, 2.0 * (lamda + qv - theta2 * beta * phi)
    L[2, 1], L[2, 2] = -kappa2_p - beta * phi, vartheta2 - 2.0 * kappa_p - 4.0 * theta * beta * phi

    if expansion_order == ExpansionOrder.SECOND:
        L[1, 3] = 3.0*qv2
        L[2, 3], L[2, 4] = 3.0 * (2.0 * qv - theta2 * beta * phi), 6.0 * qv2
        L[3, 2], L[3, 3], L[3, 4] = -2.0 * (kappa2_p + beta * phi), 3.0 * (vartheta2 - kappa_p - 2.0 * theta * beta * phi), 4.0 * (3.0 * qv - theta2 * beta * phi)
        L[4, 3], L[4, 4] = -3.0 * (kappa2_p + beta * phi), 2.0 * (vartheta2 - 2.0 * kappa_p - 4.0 * theta * beta * phi)

    # fill Hs
    H = np.zeros(n, dtype=np.complex128)
    if is_spot_measure:
        rhs = (phi * (phi + 1.0) - 2.0 * psi)
    else:
        rhs = (phi * (phi - 1.0) - 2.0 * psi)
    H[0], H[1], H[2] = 0.5 * theta2 * rhs, theta * rhs, 0.5 * rhs

    return M, L, H


@njit(cache=False, fastmath=True)
def func_rhs(t: float,   # for ode solver compatibility
             A0: np.ndarray,
             M: Tuple[np.ndarray],
             L: np.ndarray,
             H: np.ndarray
             ) -> np.ndarray:
    """
    returns rhs evaluation using matrices for the quadratic form A_t = A.T@M@A + L@A + H
    """
    n = A0.shape[0]
    quadratic = np.zeros(n, dtype=np.complex128)
    for n_ in np.arange(n):
        quadratic[n_] = A0.T @ M[n_] @ A0
    rhs = quadratic + L @ A0 + H
    return rhs


@njit(cache=False, fastmath=True)
def func_rhs_jac(t: float,   # for ode solver compatibility
                 A0: np.ndarray,
                 M: Tuple[np.ndarray],
                 L: np.ndarray,
                 H: np.ndarray
                 ) -> np.ndarray:
    """
    returns rhs jacobian evaluation using matrices for the quadratic form A_t = A.T@M@A + L@A + H
    """
    n = A0.shape[0]
    quadratic = np.zeros((n, n), dtype=np.complex128)
    for n_ in np.arange(n):
        quadratic[n_, :] = 2.0 * M[n_] @ A0
    rhs = quadratic + A0
    return rhs


# cannot use @njit(cache=False, fastmath=True) when using solve_ode_for_a with solve_ivp
def solve_ode_for_a(ttm: float,
                    theta: float,
                    kappa1: float,
                    kappa2: float,
                    beta: float,
                    volvol: float,
                    phi: np.complex128,
                    psi: np.complex128,
                    is_spot_measure: bool = True,
                    a_t0: Optional[np.ndarray] = None,
                    expansion_order: ExpansionOrder = ExpansionOrder.FIRST,
                    is_stiff_solver: bool = False,
                    dense_output: bool = False
                    ) -> OdeResult:
    """
    solve ode for given phi
    next: numba implementation to compute in range of phi
    """
    M, L, H = func_a_ode_quadratic_terms(theta=theta,
                                         kappa1=kappa1,
                                         kappa2=kappa2,
                                         beta=beta,
                                         volvol=volvol,
                                         phi=phi,
                                         psi=psi,
                                         expansion_order=expansion_order,
                                         is_spot_measure=is_spot_measure)

    if a_t0 is None:
        a_t0 = np.zeros_like(H, dtype=np.complex128)

    if is_stiff_solver:
        ode_sol = solve_ivp(fun=func_rhs, t_span=(0.0, ttm), y0=a_t0, args=(M, L, H),
                            method='BDF',
                            jac=func_rhs_jac,
                            dense_output=dense_output)
    else:
        ode_sol = solve_ivp(fun=func_rhs, t_span=(0.0, ttm), y0=a_t0, args=(M, L, H),
                            dense_output=dense_output)

    return ode_sol


@njit(cache=False, fastmath=True)
def solve_analytic_ode_for_a(ttm: float,
                             theta: float,
                             kappa1: float,
                             kappa2: float,
                             beta: float,
                             volvol: float,
                             phi: np.complex128,
                             psi: np.complex128,
                             is_spot_measure: bool,
                             a_t0: Optional[np.ndarray] = None,
                             expansion_order: ExpansionOrder = ExpansionOrder.FIRST,
                             year_days: int = 260
                             ) -> np.ndarray:
    """
    solve ode for given phi
    """
    M, L, H = func_a_ode_quadratic_terms(theta=theta,
                                         kappa1=kappa1,
                                         kappa2=kappa2,
                                         beta=beta,
                                         volvol=volvol,
                                         phi=phi,
                                         psi=psi,
                                         expansion_order=expansion_order,
                                         is_spot_measure=is_spot_measure)

    nb_steps = int(np.ceil(year_days * ttm))  # daily steps using 260 in year
    dt = ttm / nb_steps

    if a_t0 is None:
        a_t0 = np.zeros_like(H, dtype=np.complex128)
    quadratic = np.zeros_like(a_t0, dtype=np.complex128)

    w, v = la.eig(L)
    v_inv = la.inv(v)
    v_lambda = v @ np.diag(np.exp(w * dt)) @ v_inv

    reciprocal = np.reciprocal(w)
    reciprocal[0] = 0.0
    m_rhs = v @ np.diag(reciprocal*(np.exp(w * dt) - np.ones_like(H))) @ v_inv

    # fixed point
    nfp = 10
    n = H.shape[0]
    for t in np.arange(0, nb_steps):
        A_fp0 = a_t0
        for _ in np.arange(nfp):
            # for idx, m in enumerate(M):
            for n_ in np.arange(n):
                quadratic[n_] = A_fp0.T @ M[n_] @ A_fp0
            # rhs = m_rhs @ (H+quadratic)
            rhs = m_rhs @ H + quadratic * dt
            rhs[0] = (H[0] + quadratic[0])*dt
            A_fp0 = v_lambda @ a_t0 + rhs
        a_t0 = A_fp0

    return a_t0


@njit(cache=False, fastmath=True)
def solve_analytic_ode_for_a0(t_span: Tuple[float, float],
                              theta: float,
                              kappa1: float,
                              kappa2: float,
                              beta: float,
                              volvol: float,
                              phi: np.complex128,
                              psi: np.complex128,
                              expansion_order: ExpansionOrder = ExpansionOrder.FIRST
                              ) -> np.ndarray:
    """
    solve ode for given phi - too slow
    """
    M, L, H = func_a_ode_quadratic_terms(theta=theta,
                                         kappa1=kappa1,
                                         kappa2=kappa2,
                                         beta=beta,
                                         volvol=volvol,
                                         phi=phi,
                                         psi=psi,
                                         expansion_order=expansion_order)

    nt = 100
    dt = (t_span[1]-t_span[0]) / nt

    A0 = np.zeros_like(H, dtype=np.complex128)
    quadratic = np.zeros_like(A0, dtype=np.complex128)

    w, v = la.eig(L)
    v_inv = la.inv(v)
    v_lambda = v @ np.diag(np.exp(w * dt)) @ v_inv

    reciprocal = np.reciprocal(w)
    reciprocal[0] = 0.0
    m_rhs = v @ np.diag(reciprocal*(np.exp(w * dt) - np.ones_like(H))) @ v_inv

    # fixed point
    nfp = 2
    n = H.shape[0]
    Lq = np.zeros((n,n), dtype=np.complex128)

    for t in np.arange(0, nt):

        A_fp0 = A0
        for _ in np.arange(nfp):
            for n_ in np.arange(n):
                Lq[n_] = A_fp0.T @ M[n_]

            Lp = L + Lq
            w, v = la.eig(Lp)
            v_inv = la.inv(v)
            v_lambda = v @ np.diag(np.exp(w * dt)) @ v_inv

            reciprocal = np.reciprocal(w)
            reciprocal[0] = 0.0
            m_rhs = v @ np.diag(reciprocal * (np.exp(w * dt) - np.ones_like(H))) @ v_inv

            rhs = m_rhs @ H
            rhs[0] = H[0]*dt
            A_fp0 = v_lambda @ A0 + rhs

        A0 = A_fp0

    return A0


@njit(cache=False, fastmath=True)
def solve_analytic_ode_grid_phi(phi_grid: np.ndarray,
                                psi_grid: np.ndarray,
                                ttm: float,
                                theta: float,
                                kappa1: float,
                                kappa2: float,
                                beta: float,
                                volvol: float,
                                is_spot_measure: bool = True,
                                a_t0: Optional[np.ndarray] = None,
                                expansion_order: ExpansionOrder = ExpansionOrder.FIRST
                                ) -> np.ndarray:
    """
    solve ode for range phi
    next: numba implementation to compute in range of phi
    """
    if a_t0 is None:
        a_t0 = np.zeros((phi_grid.shape[0], get_expansion_n(expansion_order)), dtype=np.complex128)

    f = lambda phi, psi, a0: solve_analytic_ode_for_a(ttm=ttm,
                                                      theta=theta,
                                                      kappa1=kappa1,
                                                      kappa2=kappa2,
                                                      beta=beta,
                                                      volvol=volvol,
                                                      phi=phi,
                                                      psi=psi,
                                                      a_t0=a0,
                                                      expansion_order=expansion_order,
                                                      is_spot_measure=is_spot_measure)

    for idx, (phi, psi) in enumerate(zip(phi_grid, psi_grid)):
        a_t0[idx, :] = f(phi, psi, a_t0[idx, :])

    return a_t0


# cannot use @njit(cache=False, fastmath=True) when using solve_ode_for_a with solve_ivp
def solve_a_ode_grid(phi_grid: np.ndarray,
                     psi_grid: np.ndarray,
                     ttm: float,
                     theta: float,
                     kappa1: float,
                     kappa2: float,
                     beta: float,
                     volvol: float,
                     is_spot_measure: bool = True,
                     a_t0: Optional[np.ndarray] = None,
                     is_stiff_solver: bool = False,
                     expansion_order: ExpansionOrder = ExpansionOrder.FIRST
                     ) -> np.ndarray:
    """
    solve ode for range phi
    next: numba implementation to compute in range of phi
    """
    if a_t0 is None:
        a_t0 = np.zeros((phi_grid.shape[0], get_expansion_n(expansion_order)), dtype=np.complex128)

    f = lambda phi, psi, a0_: solve_ode_for_a(ttm=ttm,
                                              theta=theta,
                                              kappa1=kappa1,
                                              kappa2=kappa2,
                                              beta=beta,
                                              volvol=volvol,
                                              phi=phi,
                                              psi=psi,
                                              a_t0=a0_,
                                              is_stiff_solver=is_stiff_solver,
                                              dense_output=False,
                                              expansion_order=expansion_order,
                                              is_spot_measure=is_spot_measure)

    a_t1 = np.zeros((phi_grid.shape[0], get_expansion_n(expansion_order)), dtype=np.complex128)
    for idx, (phi, psi) in enumerate(zip(phi_grid, psi_grid)):
        a_t1[idx, :] = f(phi, psi, a_t0[idx, :]).y[:, -1]

    return a_t1


@njit(cache=False, fastmath=True)
def get_init_conditions_a(phi_grid: np.ndarray,
                          psi_grid: np.ndarray,
                          theta_grid: np.ndarray,
                          n_terms: int,
                          variable_type: VariableType = VariableType.LOG_RETURN
                          ) -> np.ndarray:
    """
    compute grid for a(0)
    """
    if variable_type == VariableType.LOG_RETURN:
        a_t0 = np.zeros((phi_grid.shape[0], n_terms), dtype=np.complex128)
    elif variable_type == VariableType.Q_VAR:
        a_t0 = np.zeros((psi_grid.shape[0], n_terms), dtype=np.complex128)
    elif variable_type == VariableType.SIGMA:
        a_t0 = np.zeros((theta_grid.shape[0], n_terms), dtype=np.complex128)
        a_t0[:, 1] = -theta_grid
    else:
        raise NotImplementedError
    return a_t0


def compute_logsv_a_mgf_grid(ttm: float,
                             phi_grid: np.ndarray,
                             psi_grid: np.ndarray,
                             theta_grid: np.ndarray,
                             sigma0: float,
                             theta: float,
                             kappa1: float,
                             kappa2: float,
                             beta: float,
                             volvol: float,
                             variable_type: VariableType = VariableType.LOG_RETURN,
                             expansion_order: ExpansionOrder = ExpansionOrder.SECOND,
                             a_t0: Optional[np.ndarray] = None,
                             is_stiff_solver: bool = False,
                             is_analytic: bool = False,
                             is_spot_measure: bool = True,
                             **kwargs
                             ) -> Tuple[np.ndarray, np.ndarray]:
    """
    compute
     1. solution a_t1 for ode A given a_t0
     2. log mgf function: we save an exponent calulation when pricing options
    mmg in x or QV as function of phi
    ode_solution is computed per grid of phi
    to do: numba implementation: need numba ode solvers
    """

    if a_t0 is None:
        a_t0 = get_init_conditions_a(phi_grid=phi_grid,
                                     psi_grid=psi_grid,
                                     theta_grid=theta_grid,
                                     n_terms=get_expansion_n(expansion_order=expansion_order),
                                     variable_type=variable_type)

    if is_analytic:
        a_t1 = solve_analytic_ode_grid_phi(phi_grid=phi_grid,
                                           psi_grid=psi_grid,
                                           ttm=ttm,
                                           theta=theta,
                                           kappa1=kappa1,
                                           kappa2=kappa2,
                                           beta=beta,
                                           volvol=volvol,
                                           a_t0=a_t0,
                                           expansion_order=expansion_order,
                                           is_spot_measure=is_spot_measure)
    else:
        a_t1 = solve_a_ode_grid(phi_grid=phi_grid,
                                psi_grid=psi_grid,
                                ttm=ttm,
                                theta=theta,
                                kappa1=kappa1,
                                kappa2=kappa2,
                                beta=beta,
                                volvol=volvol,
                                a_t0=a_t0,
                                is_stiff_solver=is_stiff_solver,
                                expansion_order=expansion_order,
                                is_spot_measure=is_spot_measure)

    y = sigma0 - theta
    if expansion_order == ExpansionOrder.FIRST:
        ys = np.array([1.0, y, y * y])  # params.v0 - params.theta
    elif expansion_order == ExpansionOrder.SECOND:
        y2 = y * y
        ys = np.array([1.0, y, y2, y2 * y, y2 * y2])
    else:
        raise NotImplementedError

    ys = ys + 1j * 0.0  # simple way to convert to complex array
    log_mgf = a_t1 @ ys
    return a_t1, log_mgf



"""
compute quadratic variance
"""

import numpy as np
import matplotlib.pyplot as plt
from enum import Enum

from stochvolmodels.pricers.core import mgf_pricer as mgfp
from stochvolmodels.pricers.logsv import affine_expansion as afe
from stochvolmodels.pricers.core.config import VariableType
from stochvolmodels.pricers.logsv_pricer import LogSVPricer, LogSvParams
from stochvolmodels.pricers.logsv.vol_moments_ode import compute_analytic_qvar
from stochvolmodels.pricers.core.bsm_pricer import infer_bsm_ivols_from_model_chain_prices
from stochvolmodels.utils.funcs import set_seed
from stochvolmodels.data import test_option_chain as chains
from stochvolmodels.data.option_chain import OptionChain


BTC_PARAMS = LogSvParams(sigma0=0.8327, theta=1.0139, kappa1=4.8606, kappa2=4.7938, beta=0.1985, volvol=2.3690)
VIX_PARAMS = LogSvParams(sigma0=0.9778, theta=0.5573, kappa1=4.8360, kappa2=8.6780, beta=2.3128, volvol=1.0484)
GLD_PARAMS = LogSvParams(sigma0=0.1530, theta=0.1960, kappa1=2.2068, kappa2=11.2584, beta=0.1580, volvol=2.8022)
SQQQ_PARAMS = LogSvParams(sigma0=0.9259, theta=0.9166, kappa1=3.6114, kappa2=3.9401, beta=1.1902, volvol=0.6133)
SPY_PARAMS = LogSvParams(sigma0=0.2297, theta=0.2692, kappa1=2.6949, kappa2=10.0107, beta=-1.5082, volvol=0.8503)
# BSM_PARAMS = LogSvParams(sigma0=1.0, theta=1.0, kappa1=0.0, kappa2=0.0, beta=0.0, volvol=0.0)


class UnitTests(Enum):
    QV_SLICE_PRICER = 1
    COMPARE_WITH_MC = 2


def run_unit_test(unit_test: UnitTests):

    logsv_pricer = LogSVPricer()
    params = BTC_PARAMS

    strikes = np.linspace(0.9, 2.0, 19)
    optiontypes = np.full(strikes.shape, 'C')
    expansion_order = afe.ExpansionOrder.SECOND
    variable_type = VariableType.Q_VAR
    phi_grid, psi_grid, theta_grid = mgfp.get_transform_var_grid(variable_type=variable_type, is_spot_measure=True)

    if unit_test == UnitTests.QV_SLICE_PRICER:
        ttm = 1.0
        forward = compute_analytic_qvar(params=params, ttm=ttm)
        print(forward)
        a_t1, log_mgf_grid = afe.compute_logsv_a_mgf_grid(phi_grid=phi_grid,
                                                          psi_grid=psi_grid,
                                                          theta_grid=theta_grid,
                                                          ttm=ttm,
                                                          params=params,
                                                          is_analytic=False,
                                                          expansion_order=expansion_order,
                                                          is_stiff_solver=False,
                                                          **params.to_dict()
                                                          )
        qvar_options = mgfp.slice_qvar_pricer_with_a_grid(log_mgf_grid=log_mgf_grid,
                                                          psi_grid=psi_grid,
                                                          ttm=ttm,
                                                          forward=forward,
                                                          strikes=strikes,
                                                          optiontypes=optiontypes)

        bsm_ivols = infer_bsm_ivols_from_model_chain_prices(ttms=np.array([ttm]),
                                                            forwards=np.array([forward]),
                                                            discfactors=np.array([1.0]),
                                                            strikes_ttms=(strikes,),
                                                            optiontypes_ttms=(optiontypes,),
                                                            model_prices_ttms=(qvar_options,))

        print(qvar_options)
        print(bsm_ivols)

    elif unit_test == UnitTests.COMPARE_WITH_MC:
        set_seed(24)  # 17
        option_chain = chains.get_qv_options_test_chain_data()
        option_chain = OptionChain.get_slices_as_chain(option_chain, ids=['1m', '6m'])
        fig = logsv_pricer.plot_comp_mma_inverse_options_with_mc(option_chain=option_chain,
                                                                 params=params,
                                                                 is_log_strike_xaxis=False,
                                                                 variable_type=VariableType.Q_VAR,
                                                                 is_plot_vols=False,
                                                                 nb_path=400000)

        # fig.savefig("..//..//docs//figures//model_vs_mc_qvar_logsv.pdf")

    plt.show()


if __name__ == '__main__':

    unit_test = UnitTests.QV_SLICE_PRICER

    is_run_all_tests = False
    if is_run_all_tests:
        for unit_test in UnitTests:
            run_unit_test(unit_test=unit_test)
    else:
        run_unit_test(unit_test=unit_test)



"""
key analytics for Black Scholes Merton pricer and implied volatilities
"""

import numpy as np
from numba import njit
from typing import Union, Tuple
from numba.typed import List
from scipy.stats import norm

from stochvolmodels.utils.funcs import ncdf, npdf


@njit(cache=False, fastmath=True)
def compute_normal_price(forward: float,
                         strike: float,
                         ttm: float,
                         vol: float,
                         discfactor: float = 1.0,
                         optiontype: str = 'C'
                         ) -> float:
    """
    bsm pricer for forward
    """
    sdev = forward*vol*np.sqrt(ttm)
    d = (forward - strike) / sdev
    if optiontype == 'C' or optiontype == 'IC':
        price = discfactor * ((forward-strike) * ncdf(d) + sdev * npdf(d))
    elif optiontype == 'P' or optiontype == 'IP':
        price = discfactor * ((forward - strike) * (ncdf(d)-1.0) + sdev * npdf(d))
    else:
        raise NotImplementedError(f"optiontype")

    return price


@njit(cache=False, fastmath=True)
def compute_normal_slice_prices(ttm: float,
                                forward: float,
                                strikes: np.ndarray,
                                vols: np.ndarray,
                                optiontypes: np.ndarray,
                                discfactor: float = 1.0
                                ) -> np.ndarray:
    """
    vectorised bsm deltas for array of aligned strikes, vols, and optiontypes
    """
    def f(strike: float, vol: float, optiontype: str) -> float:
        return compute_normal_price(forward=forward,
                                    ttm=ttm,
                                    vol=vol,
                                    strike=strike,
                                    optiontype=optiontype,
                                    discfactor=discfactor)
    normal_prices = np.zeros_like(strikes)
    for idx, (strike, vol, optiontype) in enumerate(zip(strikes, vols, optiontypes)):
        normal_prices[idx] = f(strike, vol, optiontype)
    return normal_prices


def compute_normal_delta_to_strike(ttm: float,
                                   forward: float,
                                   delta: float,
                                   vol: float
                                   ) -> Union[float, np.ndarray]:
    """
    bsm deltas for strikes and vols
    """
    inv_delta = norm.ppf(delta) if delta > 0.0 else norm.ppf(1.0+delta)
    sdev = forward * vol * np.sqrt(ttm)
    strike = forward - sdev*inv_delta
    return strike


@njit(cache=False, fastmath=True)
def compute_normal_delta_from_lognormal_vol(ttm: float,
                                            forward: float,
                                            strike: float,
                                            given_price: float,
                                            optiontype: str,
                                            discfactor: float = 1.0
                                            ) -> float:
    if np.abs(ttm) < 1e-12:
        if optiontype == 'C' and forward > strike:
            delta = 1.0
        elif optiontype == 'P' and forward < strike:
            delta = -1.0
        else:
            delta = 0.0
    else:
        normal_vol = infer_normal_implied_vol(forward=forward, ttm=ttm, strike=strike,
                                              given_price=given_price, optiontype=optiontype, discfactor=discfactor)
        delta = compute_normal_delta(ttm=ttm, forward=forward, strike=strike, vol=normal_vol,
                                     optiontype=optiontype, discfactor=discfactor)
    return delta


@njit(cache=False, fastmath=True)
def compute_normal_delta(ttm: float,
                         forward: float,
                         strike: float,
                         vol: float,
                         optiontype: str,
                         discfactor: float = 1.0
                         ) -> float:
    """
    bsm deltas for strikes and vols
    """
    sdev = forward * vol * np.sqrt(ttm)
    d = (forward - strike) / sdev
    if optiontype == 'C':
        normal_delta = discfactor * ncdf(d)
    elif optiontype == 'P':
        normal_delta = - discfactor * ncdf(-d)
    else:
        normal_delta = np.nan
    return normal_delta


@njit(cache=False, fastmath=True)
def compute_normal_slice_deltas(ttm: Union[float, np.ndarray],
                                forward: Union[float, np.ndarray],
                                strikes: Union[float, np.ndarray],
                                vols: Union[float, np.ndarray],
                                optiontypes: Union[np.ndarray],
                                discfactor: float = 1.0
                                ) -> Union[float, np.ndarray]:
    """
    bsm deltas for strikes and vols
    """
    sdev = forward * vols * np.sqrt(ttm)
    d = (forward - strikes) / sdev
    d1_sign = np.where(np.array([op == 'C' for op in optiontypes]), 1.0, -1.0)
    normal_deltas = discfactor * d1_sign * ncdf(d1_sign * d)
    return normal_deltas


@njit(cache=False, fastmath=True)
def compute_normal_deltas_ttms(ttms: np.ndarray,
                               forwards: np.ndarray,
                               strikes_ttms: Tuple[np.ndarray, ...],
                               vols_ttms: Tuple[np.ndarray,...],
                               optiontypes_ttms: Tuple[np.ndarray, ...],
                               ) -> List[np.ndarray]:
    """
    vectorised bsm deltas for array of aligned strikes, vols, and optiontypes
    """
    deltas_ttms = List()
    for ttm, forward, vols_ttm, strikes_ttm, optiontypes_ttm in zip(ttms, forwards, vols_ttms, strikes_ttms, optiontypes_ttms):
        deltas_ttms.append(compute_normal_slice_deltas(ttm=ttm, forward=forward, strikes=strikes_ttm, vols=vols_ttm, optiontypes=optiontypes_ttm))
    return deltas_ttms


@njit(cache=False, fastmath=True)
def compute_normal_slice_vegas(ttm: float,
                               forward: float,
                               strikes: np.ndarray,
                               vols: np.ndarray,
                               optiontypes: np.ndarray = None
                               ) -> np.ndarray:
    """
    vectorised bsm vegas for array of aligned strikes, vols, and optiontypes
    """
    sdev = forward*vols * np.sqrt(ttm)
    d = (forward - strikes) / sdev
    vegas = forward * npdf(d) * np.sqrt(ttm)
    return vegas


@njit(cache=False, fastmath=True)
def compute_normal_vegas_ttms(ttms: np.ndarray,
                              forwards: np.ndarray,
                              strikes_ttms: Tuple[np.ndarray, ...],
                              vols_ttms: Tuple[np.ndarray,...],
                              optiontypes_ttms: Tuple[np.ndarray, ...],
                              ) -> List[np.ndarray]:
    """
    vectorised bsm vegas for array of aligned strikes, vols, and optiontypes
    """
    vegas_ttms = List()
    for ttm, forward, vols_ttm, strikes_ttm, optiontypes_ttm in zip(ttms, forwards, vols_ttms, strikes_ttms, optiontypes_ttms):
        vegas_ttms.append(compute_normal_slice_vegas(ttm=ttm, forward=forward, strikes=strikes_ttm, vols=vols_ttm, optiontypes=optiontypes_ttm))
    return vegas_ttms


@njit(cache=False, fastmath=True)
def infer_normal_implied_vol(forward: float,
                             ttm: float,
                             strike: float,
                             given_price: float,
                             discfactor: float = 1.0,
                             optiontype: str = 'C',
                             tol: float = 1e-12,
                             is_bounds_to_nan: bool = False
                             ) -> float:
    """
    compute normal implied vol
    """
    x1, x2 = 0.01, 10.0  # starting values
    f = compute_normal_price(forward=forward, strike=strike, ttm=ttm, vol=x1, discfactor=discfactor, optiontype=optiontype) - given_price
    fmid = compute_normal_price(forward=forward, strike=strike, ttm=ttm, vol=x2, discfactor=discfactor, optiontype=optiontype) - given_price

    if f*fmid < 0.0:
        if f < 0.0:
            rtb = x1
            dx = x2-x1
        else:
            rtb = x2
            dx = x1-x2
        xmid = rtb
        for j in range(0, 100):
            dx = dx*0.5
            xmid = rtb+dx
            fmid = compute_normal_price(forward=forward, strike=strike, ttm=ttm, vol=xmid, discfactor=discfactor, optiontype=optiontype) - given_price
            if fmid <= 0.0:
                rtb = xmid
            if np.abs(fmid) < tol:
                break
        v1 = xmid

    else:
        if f < 0:
            v1 = x1
        else:
            v1 = x2

    if is_bounds_to_nan:  # in case vol was inferred it will return nan
        if np.abs(v1-x1) < tol or np.abs(v1-x2) < tol:
            v1 = np.nan
    return v1


@njit(cache=False, fastmath=True)
def infer_normal_ivols_from_model_slice_prices(ttm: float,
                                               forward: float,
                                               strikes: np.ndarray,
                                               optiontypes: np.ndarray,
                                               model_prices: np.ndarray,
                                               discfactor: float
                                               ) -> np.ndarray:
    model_vol_ttm = np.zeros_like(strikes)
    for idx, (strike, model_price, optiontype) in enumerate(zip(strikes, model_prices, optiontypes)):
        model_vol_ttm[idx] = infer_normal_implied_vol(forward=forward, ttm=ttm, discfactor=discfactor,
                                                      given_price=model_price,
                                                      strike=strike,
                                                      optiontype=optiontype)
    return model_vol_ttm


@njit(cache=False, fastmath=True)
def infer_normal_ivols_from_slice_prices(ttm: float,
                                         forward: float,
                                         discfactor: float,
                                         strikes: np.ndarray,
                                         optiontypes: np.ndarray,
                                         model_prices: np.ndarray
                                         ) -> List:
    """
    vectorised chain ivols
    """
    model_vol_ttm = np.zeros_like(strikes)
    for idx, (strike, model_price, optiontype) in enumerate(zip(strikes, model_prices, optiontypes)):
        model_vol_ttm[idx] = infer_normal_implied_vol(forward=forward, ttm=ttm, discfactor=discfactor,
                                                      given_price=model_price,
                                                      strike=strike,
                                                      optiontype=optiontype)
    return model_vol_ttm


@njit(cache=False, fastmath=True)
def infer_normal_ivols_from_chain_prices(ttms: np.ndarray,
                                         forwards: np.ndarray,
                                         discfactors: np.ndarray,
                                         strikes_ttms: List[np.ndarray,...],
                                         optiontypes_ttms: List[np.ndarray, ...],
                                         model_prices_ttms: List[np.ndarray],
                                         ) -> List[np.ndarray, ...]:
    """
    vectorised chain ivols
    """
    model_vol_ttms = List()
    for ttm, forward, discfactor, strikes_ttm, optiontypes_ttm, model_prices_ttm in zip(ttms, forwards, discfactors, strikes_ttms, optiontypes_ttms, model_prices_ttms):
        model_vol_ttm = np.zeros_like(strikes_ttm)
        for idx, (strike, model_price, optiontype) in enumerate(zip(strikes_ttm, model_prices_ttm, optiontypes_ttm)):
            model_vol_ttm[idx] = infer_normal_implied_vol(forward=forward, ttm=ttm, discfactor=discfactor,
                                                          given_price=model_price,
                                                          strike=strike,
                                                          optiontype=optiontype)
        model_vol_ttms.append(model_vol_ttm)
    return model_vol_ttms



"""
Montecarlo analytics for option pay-off computations
"""

import numpy as np
from numba import njit

from ...pricers.core.config import VariableType


@njit(cache=False, fastmath=True)
def compute_mc_vars_payoff(x0: np.ndarray, sigma0: np.ndarray, qvar0: np.ndarray,
                           ttm: float,
                           forward: float,
                           strikes_ttm: np.ndarray,
                           optiontypes_ttm: np.ndarray,
                           discfactor: float = 1.0,
                           variable_type: VariableType = VariableType.LOG_RETURN
                           ) -> (np.ndarray, np.ndarray):

    # need to remember it for options on QVAR
    spots_t = forward*np.exp(x0)
    correnction = np.nanmean(spots_t) - forward
    spots_t = spots_t - correnction

    if variable_type == VariableType.LOG_RETURN:
        underlying_t = spots_t
    elif variable_type == VariableType.Q_VAR:
        underlying_t = qvar0 / ttm
    else:
        raise NotImplementedError

    option_prices = np.zeros_like(strikes_ttm)
    option_std = np.zeros_like(strikes_ttm)
    for idx, (strike, type_) in enumerate(zip(strikes_ttm, optiontypes_ttm)):
        if type_ == 'C':
            payoff = np.where(np.greater(underlying_t, strike), underlying_t-strike, 0.0)
        elif type_ == 'IC':
            payoff = np.where(np.greater(underlying_t, strike), underlying_t - strike, 0.0) / spots_t
        elif type_ == 'P':
            payoff = np.where(np.less(underlying_t, strike), strike-underlying_t, 0.0)
        elif type_ == 'IP':
            payoff = np.where(np.less(underlying_t, strike), strike - underlying_t, 0.0) / spots_t
        else:
            payoff = np.zeros_like(underlying_t)
        option_prices[idx] = discfactor*np.nanmean(payoff)
        option_std[idx] = discfactor*np.nanstd(payoff)

    return option_prices, option_std/np.sqrt(x0.shape[0])



"""
test option valuation using moment generating function analytics for Black-Scholes-Merton model
"""

import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import seaborn as sns
from typing import Tuple
from enum import Enum

from stochvolmodels.pricers.core import mgf_pricer as mgfp
from stochvolmodels.pricers.core.bsm_pricer import infer_bsm_ivols_from_model_chain_prices
from stochvolmodels.pricers.core.config import VariableType


def compute_normal_mgf_grid(ttm: float,
                            vol: float,
                            is_spot_measure: bool = True
                            ) -> Tuple[np.ndarray, np.ndarray]:
    """
    for normal: mgf = exp( 0.5*PHI * (PHI + 1.0)*sigma^2*T )
    add two columns to make compatible with @ for numba in slice_pricer_with_a_grid
    """
    phi_grid = mgfp.get_phi_grid(is_spot_measure=is_spot_measure)
    if is_spot_measure:  # use 1 for spot measure
        alpha = 1.0
    else:
        alpha = -1.0
    log_mgf_grid = 0.5 * phi_grid * (phi_grid + alpha) * (ttm * vol * vol)
    return log_mgf_grid, phi_grid


def compute_normal_mgf_psi_grid(ttm: float,
                                vol: float,
                                is_spot_measure: bool = True
                                ) -> Tuple[np.ndarray, np.ndarray]:
    psi_grid = mgfp.get_psi_grid(is_spot_measure=is_spot_measure)
    log_mgf_grid = -psi_grid * (ttm * vol * vol)
    return log_mgf_grid, psi_grid


def bsm_slice_pricer(ttm: float,
                     forward: float,
                     vol: float,
                     strikes: np.ndarray,
                     optiontypes: np.ndarray,
                     variable_type: VariableType = VariableType.LOG_RETURN,
                     is_spot_measure: bool = True
                     ) -> Tuple[np.ndarray, np.ndarray]:

    if variable_type == VariableType.LOG_RETURN:
        log_mgf_grid, phi_grid = compute_normal_mgf_grid(ttm=ttm, vol=vol, is_spot_measure=is_spot_measure)
        bsm_prices = mgfp.slice_pricer_with_mgf_grid(log_mgf_grid=log_mgf_grid,
                                                     phi_grid=phi_grid,
                                                     ttm=ttm,
                                                     forward=forward,
                                                     strikes=strikes,
                                                     optiontypes=optiontypes,
                                                     is_spot_measure=is_spot_measure)
        bsm_ivols = infer_bsm_ivols_from_model_chain_prices(ttms=np.array([ttm]),
                                                            forwards=np.array([forward]),
                                                            discfactors=np.array([1.0]),
                                                            strikes_ttms=(strikes,),
                                                            optiontypes_ttms=(optiontypes,),
                                                            model_prices_ttms=(bsm_prices,))
    elif variable_type == VariableType.Q_VAR:
        log_mgf_grid, psi_grid = compute_normal_mgf_psi_grid(ttm=ttm, vol=vol, is_spot_measure=is_spot_measure)
        bsm_prices = mgfp.slice_qvar_pricer_with_a_grid(log_mgf_grid=log_mgf_grid,
                                                        psi_grid=psi_grid,
                                                        ttm=ttm,
                                                        strikes=strikes,
                                                        optiontypes=optiontypes,
                                                        is_spot_measure=is_spot_measure,
                                                        forward=forward)
        bsm_ivols = np.zeros_like(bsm_prices)
    else:
        raise ValueError(f"not implemented")

    return bsm_prices, bsm_ivols


def compare_spot_and_inverse_options():
    ttm = 1.0
    forward = 1.0
    vol = 1.0
    strikes = np.linspace(0.5, 5.0, 19)
    optiontypes = np.full(strikes.shape, 'C')
    optiontypes_inverse = np.full(strikes.shape, 'IC')

    # spot
    bsm_prices_spot, bsm_ivols = bsm_slice_pricer(ttm=ttm, forward=forward, vol=vol, strikes=strikes,
                                                  optiontypes=optiontypes,
                                                  is_spot_measure=True)
    bsm_prices_inverse, bsm_ivols = bsm_slice_pricer(ttm=ttm, forward=forward, vol=vol, strikes=strikes,
                                                     optiontypes=optiontypes_inverse,
                                                     is_spot_measure=False)

    bsm_prices_spot = pd.Series(bsm_prices_spot, index=strikes, name='spot')
    bsm_prices_inverse = pd.Series(bsm_prices_inverse, index=strikes, name='inverse')

    prices = pd.concat([bsm_prices_spot, bsm_prices_inverse], axis=1)
    with sns.axes_style("darkgrid"):
        fig, ax = plt.subplots(1, 1, figsize=(10, 4.0), tight_layout=True)
        sns.lineplot(data=prices, ax=ax)


def compare_spot_and_inverse_qvar_options():
    ttm = 1.0
    forward = 1.0
    vol = 1.0
    strikes = np.linspace(0.5, 5.0, 19)
    optiontypes = np.full(strikes.shape, 'C')
    optiontypes_inverse = np.full(strikes.shape, 'IC')

    # spot
    bsm_prices_spot, bsm_ivols = bsm_slice_pricer(ttm=ttm, forward=forward, vol=vol, strikes=strikes,
                                                  optiontypes=optiontypes,
                                                  variable_type=VariableType.Q_VAR,
                                                  is_spot_measure=True)
    bsm_prices_inverse, bsm_ivols = bsm_slice_pricer(ttm=ttm, forward=forward, vol=vol, strikes=strikes,
                                                     optiontypes=optiontypes_inverse,
                                                     variable_type=VariableType.Q_VAR,
                                                     is_spot_measure=False)

    bsm_prices_spot = pd.Series(bsm_prices_spot, index=strikes, name='spot')
    bsm_prices_inverse = pd.Series(bsm_prices_inverse, index=strikes, name='inverse')

    prices = pd.concat([bsm_prices_spot, bsm_prices_inverse], axis=1)
    with sns.axes_style("darkgrid"):
        fig, ax = plt.subplots(1, 1, figsize=(10, 4.0), tight_layout=True)
        sns.lineplot(data=prices, ax=ax)


class UnitTests(Enum):
    BSM_SLICE_PRICER = 1
    SPOT_INVERSE_COMP = 2
    SPOT_INVERSE_QVAR_COMP = 3


def run_unit_test(unit_test: UnitTests):

    if unit_test == UnitTests.BSM_SLICE_PRICER:
        ttm = 1.0
        forward = 1.0
        vol = 1.0
        strikes = np.linspace(0.5, 5.0, 19)
        optiontypes = np.full(strikes.shape, 'C')
        bsm_prices, bsm_ivols = bsm_slice_pricer(ttm=ttm, forward=forward, vol=vol, strikes=strikes, optiontypes=optiontypes)
        print(bsm_prices)
        print(bsm_ivols)

    elif unit_test == UnitTests.SPOT_INVERSE_COMP:
        compare_spot_and_inverse_options()

    elif unit_test == UnitTests.SPOT_INVERSE_QVAR_COMP:
        compare_spot_and_inverse_qvar_options()

    plt.show()


if __name__ == '__main__':

    unit_test = UnitTests.BSM_SLICE_PRICER

    is_run_all_tests = False
    if is_run_all_tests:
        for unit_test in UnitTests:
            run_unit_test(unit_test=unit_test)
    else:
        run_unit_test(unit_test=unit_test)



"""
data container for option chain
data is provided as:
1 arrays of ttms, forwards, discounts
2 lists of arrays with strikes, optiom types and bid / ask prices and vols
"""

from __future__ import annotations

import numpy as np
from dataclasses import dataclass
from typing import Tuple, Optional
from numba.typed import List

import stochvolmodels.pricers.core.bsm_pricer as bsm


@dataclass
class OptionSlice:
    """
    container for slice data
    """
    ttm: float
    forward: float
    strikes: np.ndarray
    optiontypes: np.ndarray
    id: str
    discfactor: float = None  # discount factors
    discount_rate: float = None  # discount rates
    bid_ivs: Optional[np.ndarray] = None
    ask_ivs: Optional[np.ndarray] = None
    bid_prices: Optional[np.ndarray] = None
    ask_prices: Optional[np.ndarray] = None

    def __post_init__(self):
        """
        to do: check dimension aligmnent
        make consistent discfactors
        """
        if self.discfactor is not None:
            self.discount_rate = - np.log(self.discfactor) / self.ttm
        elif self.discount_rate is not None:
            self.discfactor = np.exp(-self.discount_rate * self.ttm)
        else:  # use zeros
            self.discfactor = 1.0
            self.discount_rate = 0.0


@dataclass
class OptionChain:
    """
    container for chain data
    note we do not use chain as list of slices here
    for extensive use of numba we use List[np.ndarray] with per slice data
    """
    ttms: np.ndarray
    forwards: np.ndarray
    strikes_ttms: List[np.ndarray]
    optiontypes_ttms: List[np.ndarray]
    ids: Optional[np.ndarray]  # slice_t names
    discfactors: Optional[np.ndarray] = None  # discount factors
    discount_rates: Optional[np.ndarray] = None  # discount rates
    ticker: Optional[str] = None  # associated ticker
    bid_ivs: Optional[List[np.ndarray]] = None
    ask_ivs: Optional[List[np.ndarray]] = None
    bid_prices: Optional[List[np.ndarray]] = None
    ask_prices: Optional[List[np.ndarray]] = None

    def __post_init__(self):
        """
        to do: check dimension aligmnent
        make consistent discfactors
        """
        if self.discfactors is not None:
            self.discount_rates = - np.log(self.discfactors) / self.ttms
        elif self.discount_rates is not None:
            self.discfactors = np.exp(-self.discount_rates * self.ttms)
        else:  # use zeros
            self.discfactors = np.ones_like(self.ttms)
            self.discount_rates = np.zeros_like(self.ttms)

    @classmethod
    def slice_to_chain(cls,
                       ttm: float,
                       forward: float,
                       strikes: np.ndarray,
                       optiontypes: np.ndarray,
                       discfactor: float = 1.0,
                       id: Optional[str] = None
                       ) -> OptionChain:

        return cls(ttms=np.array([ttm]),
                   forwards=np.array([forward]),
                   strikes_ttms=(strikes,),
                   optiontypes_ttms=(optiontypes,),
                   discfactors=np.array([discfactor]),
                   ids=np.array([id]) if id is not None else np.array([f"{ttm:0.2f}"]))

    def get_mid_vols(self) -> List[np.ndarray]:
        if self.bid_ivs is not None and self.ask_ivs is not None:
            return List(0.5 * (bid_iv + ask_iv) for bid_iv, ask_iv in zip(self.bid_ivs, self.ask_ivs))
        else:
            return None

    def get_chain_deltas(self) -> List[np.ndarray]:
        deltas_ttms = bsm.compute_bsm_deltas_ttms(ttms=self.ttms,
                                                  forwards=self.forwards,
                                                  strikes_ttms=self.strikes_ttms,
                                                  optiontypes_ttms=self.optiontypes_ttms,
                                                  vols_ttms=self.get_mid_vols())
        return deltas_ttms

    def get_chain_vegas(self, is_unit_ttm_vega: bool = False) -> List[np.ndarray]:
        if is_unit_ttm_vega:
            ttms = np.ones_like(self.ttms)
        else:
            ttms = self.ttms
        vegas_ttms = bsm.compute_bsm_vegas_ttms(ttms=ttms,
                                                forwards=self.forwards,
                                                strikes_ttms=self.strikes_ttms,
                                                optiontypes_ttms=self.optiontypes_ttms,
                                                vols_ttms=self.get_mid_vols())
        return vegas_ttms

    def get_chain_atm_vols(self) -> np.ndarray:
        atm_vols = np.zeros(len(self.ttms))
        for idx, (forward, strikes_ttm, y) in enumerate(zip(self.forwards, self.strikes_ttms, self.get_mid_vols())):
            atm_vols[idx] = np.interp(x=forward, xp=strikes_ttm, fp=y)
        return atm_vols

    def get_chain_data_as_xy(self) -> Tuple[List[np.ndarray], np.ndarray]:
        """
        these data are needed for to pass x and y for model calibrations
        """
        mid_vols = List(0.5 * (bid_iv + ask_iv) for bid_iv, ask_iv in zip(self.bid_ivs, self.ask_ivs))
        x = (self.ttms, self.forwards, self.discfactors, self.strikes_ttms, self.optiontypes_ttms)
        y = mid_vols
        return x, y

    def compute_model_ivols_from_chain_data(self, model_prices: List[np.ndarray]) -> List[np.ndarray]:
        model_ivols = bsm.infer_bsm_ivols_from_model_chain_prices(ttms=self.ttms,
                                                                  forwards=self.forwards,
                                                                  discfactors=self.discfactors,
                                                                  strikes_ttms=self.strikes_ttms,
                                                                  optiontypes_ttms=self.optiontypes_ttms,
                                                                  model_prices_ttms=model_prices)
        return model_ivols

    @classmethod
    def to_uniform_strikes(cls, obj, num_strikes=21):
        """
        in some situations (like model price display) we want to get a uniform grid corresponding to the chain
        bid_ivs and ask_ivs will be set to none
        """
        new_strikes_ttms = List()
        new_optiontypes_ttms = List()
        for strikes_ttm, forward in zip(obj.strikes_ttms, obj.forwards):
            new_strikes = np.linspace(strikes_ttm[0], strikes_ttm[-1], num_strikes)
            new_strikes_ttms.append(new_strikes)
            new_optiontypes_ttms.append(np.where(new_strikes >= forward, 'C', 'P'))

        return cls(ttms=obj.ttms, forwards=obj.forwards, strikes_ttms=new_strikes_ttms,
                   optiontypes_ttms=new_optiontypes_ttms, discfactors=obj.discfactors,
                   ticker=obj.ticker,
                   ids=obj.ids,
                   bid_ivs=None, ask_ivs=None)

    def get_slice(self, id: str) -> OptionSlice:
        idx = list(self.ids).index(id)
        option_slice = OptionSlice(id=self.ids[idx],
                                   ttm=self.ttms[idx],
                                   forward=self.forwards[idx],
                                   strikes=self.strikes_ttms[idx],
                                   optiontypes=self.optiontypes_ttms[idx],
                                   discfactor=self.discfactors[idx],
                                   bid_ivs=None if self.bid_ivs is None else self.bid_ivs[idx],
                                   ask_ivs=None if self.ask_ivs is None else self.ask_ivs[idx],
                                   bid_prices=None if self.bid_prices is None else self.bid_prices[idx],
                                   ask_prices=None if self.ask_prices is None else self.ask_prices[idx])
        return option_slice

    @classmethod
    def get_slices_as_chain(cls, option_chain: OptionChain, ids: List[str]) -> OptionChain:
        indices = np.in1d(option_chain.ids, ids).nonzero()[0]
        option_chain = cls(ids=ids,
                         ttms=option_chain.ttms[indices],
                         ticker=option_chain.ticker,
                         forwards=option_chain.forwards[indices],
                         strikes_ttms=List(option_chain.strikes_ttms[idx] for idx in indices),
                         optiontypes_ttms=List(option_chain.optiontypes_ttms[idx] for idx in indices),
                         discfactors=option_chain.discfactors[indices],
                         bid_ivs=None if option_chain.bid_ivs is None else List(option_chain.bid_ivs[idx] for idx in indices),
                         ask_ivs=None if option_chain.ask_ivs is None else List(option_chain.ask_ivs[idx] for idx in indices),
                         bid_prices=None if option_chain.bid_prices is None else List(option_chain.bid_prices[idx] for idx in indices),
                         ask_prices=None if option_chain.ask_prices is None else List(option_chain.ask_prices[idx] for idx in indices))
        return option_chain

    @classmethod
    def get_uniform_chain(cls,
                          ttms: np.ndarray = np.array([0.083, 0.25]),
                          ids: np.ndarray = np.array(['1m', '3m']),
                          # strikes: np.ndarray = np.linspace(0.5, 1.5, 11)
                          strikes: np.ndarray = 100.0*np.linspace(0.9, 1.1, 3),
                          flat_vol: float = 0.2
                          ) -> OptionChain:
        return cls(ttms=ttms,
                   ids=ids,
                   forwards=100.0*np.ones_like(ttms),
                   strikes_ttms=List([strikes for _ in ttms]),
                   bid_ivs=List([flat_vol*np.ones_like(strikes) for _ in ttms]),
                   ask_ivs=List([flat_vol*np.ones_like(strikes) for _ in ttms]),
                   optiontypes_ttms=List([np.where(strikes >= 1.0, 'C', 'P') for _ in ttms]))


"""
analytics for vol and QV moments computation
"""

import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import seaborn as sns
from numpy import linalg as la
from scipy import linalg as sla
from enum import Enum

from ...pricers.logsv_pricer import LogSvParams
from ...utils.funcs import set_seed


VOLVOL = 1.75

DRIFT_PARAMS = {'$(\kappa_{1}=4, \kappa_{2}=0)$': LogSvParams(sigma0=1.0, theta=1.0, kappa1=4.0, kappa2=0.0, beta=0.0, volvol=VOLVOL),
                '$(\kappa_{1}=4, \kappa_{2}=4)$': LogSvParams(sigma0=1.0, theta=1.0, kappa1=4.0, kappa2=4.0, beta=0.0, volvol=VOLVOL),
                '$(\kappa_{1}=4, \kappa_{2}=8)$': LogSvParams(sigma0=1.0, theta=1.0, kappa1=4.0, kappa2=8.0, beta=0.0, volvol=VOLVOL)}


def compute_analytic_vol_moments(params: LogSvParams,
                                 t: float = 1.0,
                                 n_terms: int = 4,
                                 is_qvar: bool = False
                                 ) -> np.ndarray:

    lambda_m = params.get_vol_moments_lambda(n_terms=n_terms)

    y = params.sigma0 - params.theta
    y0 = np.zeros(n_terms)
    for n in range(0, n_terms):
        y0[n] = np.power(y, n+1)

    if np.isclose(np.abs(t), 0.0):
        return y0

    rhs = np.zeros(n_terms)
    rhs[1] = params.vartheta2*params.theta2

    if is_qvar: # need flat boundary condition
        rhs[-1] = -n_terms*params.kappa2*np.power(y, n_terms+1)
    else:
        rhs[-1] = -n_terms*params.kappa2*np.power(y, n_terms+1)

    i_m = la.inv(lambda_m)
    is_expm = True
    if is_expm:
        e_m = sla.expm(lambda_m*t)
        m_rhs = i_m @ (e_m - np.eye(n_terms))
    else:
        w, v = la.eig(lambda_m)
        v_inv = la.inv(v)
        e_m = np.real(v @ np.diag(np.exp(w*t)) @ v_inv)
        m_rhs = np.real(v @ np.diag(np.reciprocal(w)*(np.exp(w*t)-np.ones(n_terms))) @ v_inv)
        # m_rhs = i_m @ (e_m - np.eye(n_terms))

    if is_qvar:
        sol1 = m_rhs @ y0
        intm2 = i_m @ (m_rhs-t*np.eye(n_terms))
        sol2 = intm2 @ rhs
    else:
        sol1 = e_m @ y0
        sol2 = m_rhs @ rhs

    sol = sol1 + sol2
    return sol


def compute_analytic_qvar(params: LogSvParams,
                          ttm: float = 1.0,
                          n_terms: int = 4
                          ) -> float:
    if np.isclose(ttm, 0.0):
        qvar = np.square(params.sigma0)
    else:
        int_moments = compute_analytic_vol_moments(params=params, t=ttm, n_terms=n_terms, is_qvar=True)
        qvar = (int_moments[1] + 2.0*params.theta*int_moments[0]) / ttm + params.theta2
    return qvar


def compute_vol_moments_t(params: LogSvParams,
                          ttm: np.ndarray,
                          n_terms: int = 4,
                          is_print: bool = False
                          ) -> np.ndarray:
    moments = np.zeros((len(ttm), n_terms))
    for idx, t_ in enumerate(ttm):
        moments_ = compute_analytic_vol_moments(t=t_, params=params, n_terms=n_terms)
        if is_print:
            print(f"t={t_}: {moments_}")
        moments[idx, :] = moments_
    return moments


def compute_expected_vol_t(params: LogSvParams,
                           t: np.ndarray,
                           n_terms: int = 4,
                           ) -> np.ndarray:
    ev = np.zeros(len(t))
    for idx, t_ in enumerate(t):
        moments = compute_analytic_vol_moments(t=t_, params=params, n_terms=n_terms)
        ev[idx] = moments[0] + params.theta
    return ev


def compute_sqrt_qvar_t(params: LogSvParams, t: np.ndarray, n_terms: int = 4) -> np.ndarray:
    ev = np.zeros(len(t))
    for idx, t_ in enumerate(t):
        ev[idx] = np.sqrt(compute_analytic_qvar(ttm=t_, params=params, n_terms=n_terms))
    return ev


class UnitTests(Enum):
    VOL_MOMENTS = 1
    EXPECTED_VOL = 2
    EXPECTED_QVAR = 3


def run_unit_test(unit_test: UnitTests):

    from stochvolmodels.pricers.logsv_pricer import LogSVPricer
    logsv_pricer = LogSVPricer()

    n_terms = 4
    nb_path = 200000
    ttm = 1.0
    params = LogSvParams(sigma0=1.0, theta=1.0, kappa1=4.0, kappa2=4.0, beta=0.0, volvol=1.75)
    params.assert_vol_moments_stability(n_terms=n_terms)
    set_seed(8)  # 8
    sigma_t, grid_t = logsv_pricer.simulate_vol_paths(ttm=ttm, params=params, nb_path=nb_path)

    if unit_test == UnitTests.VOL_MOMENTS:

        mcs = []
        for n in np.arange(n_terms):
            if n > 0:
                m_n = np.power(sigma_t-params.theta, n+1)
            else:
                m_n = sigma_t - params.theta
            mc_mean, mc_std = np.mean(m_n, axis=1), np.std(sigma_t, axis=1) / np.sqrt(nb_path)
            mc = pd.Series(mc_mean, index=grid_t, name=f"MC m{n+1}")
            mc_m = pd.Series(mc_mean-1.96*mc_std, index=grid_t, name='MC-cd')
            mc_p = pd.Series(mc_mean+1.96*mc_std, index=grid_t, name='MC+cd')
            mcs.append(mc)
        analytic_vol_moments = compute_vol_moments_t(params=params, ttm=grid_t, n_terms=n_terms)
        analytic_vol_moments = pd.DataFrame(analytic_vol_moments, index=grid_t, columns=[f"m{n+1}" for n in range(n_terms)])
        mcs = pd.concat(mcs, axis=1)

        df = pd.concat([analytic_vol_moments, mcs], axis=1)
        print(df)
        df.plot()

    elif unit_test == UnitTests.EXPECTED_VOL:

        mc_mean, mc_std = np.mean(sigma_t, axis=1), np.std(sigma_t, axis=1) / np.sqrt(nb_path)
        mc = pd.Series(mc_mean, index=grid_t, name='MC')
        mc_m = pd.Series(mc_mean-1.96*mc_std, index=grid_t, name='MC-cd')
        mc_p = pd.Series(mc_mean+1.96*mc_std, index=grid_t, name='MC+cd')

        analytic_vol_moments = compute_expected_vol_t(params=params, t=grid_t, n_terms=n_terms)
        analytic_vol_moments = pd.Series(analytic_vol_moments, index=grid_t, name='Analytic')

        df = pd.concat([analytic_vol_moments, mc, mc_m, mc_p], axis=1)
        print(df)
        df.plot()

    elif unit_test == UnitTests.EXPECTED_QVAR:

        q_var = pd.DataFrame(np.square(sigma_t)).expanding(axis=0).mean().to_numpy()
        mc_mean = np.sqrt(np.mean(q_var, axis=1))
        mc_std = np.std(q_var, axis=1) / np.sqrt(nb_path)
        mc = pd.Series(mc_mean, index=grid_t, name='MC')
        mc_m = pd.Series(mc_mean-1.96*mc_std, index=grid_t, name='MC-cd')
        mc_p = pd.Series(mc_mean+1.96*mc_std, index=grid_t, name='MC+cd')

        analytic_vol_moments = compute_sqrt_qvar_t(params=params, t=grid_t, n_terms=n_terms)
        analytic_vol_moments = pd.Series(analytic_vol_moments, index=grid_t, name='Analytic')

        df = pd.concat([analytic_vol_moments, mc, mc_m, mc_p], axis=1)
        with sns.axes_style('darkgrid'):
            fig, ax = plt.subplots(1, 1, figsize=(18, 10), tight_layout=True)
            sns.lineplot(data=analytic_vol_moments, dashes=False, ax=ax)
            ax.errorbar(x=df.index[::5], y=mc_mean[::5], yerr=mc_std[::5], fmt='o', color='green', capsize=8)

    plt.show()


if __name__ == '__main__':

    unit_test = UnitTests.VOL_MOMENTS

    is_run_all_tests = False
    if is_run_all_tests:
        for unit_test in UnitTests:
            run_unit_test(unit_test=unit_test)
    else:
        run_unit_test(unit_test=unit_test)
